from typing import Tuple, List, Optional, Dict, cast

import numpy as np
from numpy.typing import NDArray
from deepxube.base.nnet_input import FlatIn, StateGoalIn
from deepxube.base.domain import State, Action, Goal, ActsEnumFixed, StartGoalWalkable, StateGoalVizable, StringToAct
from deepxube.factories.domain_factory import domain_factory
from deepxube.factories.nnet_input_factory import register_nnet_input

from matplotlib.figure import Figure

import pickle

import pathlib
import tarfile
import os
import wget  # type: ignore
from filelock import FileLock


class SkState(State):
    __slots__ = ['agent', 'boxes', 'walls', 'hash']

    def __init__(self, agent: NDArray[np.int_], boxes: NDArray[np.uint8], walls: NDArray[np.uint8]):
        self.agent: NDArray[np.int_] = agent
        self.boxes: NDArray[np.uint8] = boxes
        self.walls: NDArray[np.uint8] = walls

        self.hash: Optional[int] = None

    def __hash__(self) -> int:
        if self.hash is None:
            boxes_flat = self.boxes.flatten()
            walls_flat = self.walls.flatten()
            state: NDArray[np.int_] = np.concatenate((self.agent, boxes_flat.astype(int), walls_flat.astype(int)), axis=0)

            self.hash = hash(state.tobytes())

        return self.hash

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SkState):
            agents_eq: bool = np.array_equal(self.agent, other.agent)
            boxes_eq: bool = np.array_equal(self.boxes, other.boxes)
            walls_eq: bool = np.array_equal(self.walls, other.walls)

            return agents_eq and boxes_eq and walls_eq
        return NotImplemented


class SkGoal(Goal):
    __slots__ = ['boxes']

    def __init__(self, boxes: NDArray[np.uint8]):
        self.boxes: NDArray[np.uint8] = boxes


class SkAction(Action):
    def __init__(self, action: int):
        self.action = action

    def __hash__(self) -> int:
        return self.action

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SkAction):
            return self.action == other.action
        return NotImplemented


def load_states(data_dir: str) -> List[SkState]:
    states_np = pickle.load(open(f"{data_dir}/sokoban/train.pkl", "rb"))
    # t_file = tarfile.open(file_name, "r:gz")
    # states_np = pickle.load(t_file.extractfile(t_file.getmembers()[1]))

    states: List[SkState] = []

    agent_idxs: Tuple[NDArray, ...] = np.where(states_np == 1)
    box_masks: NDArray = np.array(states_np == 2)
    wall_masks: NDArray = np.array(states_np == 4)

    idx: int
    for idx in range(states_np.shape[0]):
        agent_idx = np.array([agent_idxs[1][idx], agent_idxs[2][idx]], dtype=int)

        states.append(SkState(agent_idx, box_masks[idx], wall_masks[idx]))

    return states


def _get_surfaces() -> Dict[str, NDArray]:
    import imageio.v2 as imageio
    data_dir = get_data_dir()
    img_dir = f"{data_dir}/sokoban/"

    lock = FileLock(f"{data_dir}/file.lock")
    with lock:
        # Load images, representing the corresponding situation
        surface_dict: Dict[str, NDArray] = dict()
        for img_name in ["box", "box_on_target", "box_target", "floor", "player", "player_on_target", "wall"]:
            surface_dict[img_name] = imageio.imread(f"{img_dir}/surface/{img_name}.png")

    return surface_dict


def _get_train_states() -> List[SkState]:
    data_dir = get_data_dir()
    lock = FileLock(f"{data_dir}/file.lock")

    with lock:
        states_train: List[SkState] = load_states(data_dir)

    return states_train


def get_data_dir() -> str:
    parent_dir: str = str(pathlib.Path(__file__).parent.resolve())
    data_dir: str = f"{parent_dir}/data/sokoban/"

    return data_dir


@domain_factory.register_class("sokoban")
class Sokoban(ActsEnumFixed[SkState, SkAction, SkGoal], StartGoalWalkable[SkState, SkAction, SkGoal], StateGoalVizable[SkState, SkAction, SkGoal],
              StringToAct[SkState, SkAction, SkGoal]):
    def __init__(self) -> None:
        super().__init__()

        self.dim: int = 10
        self.num_boxes: int = 4

        self.num_actions: int = 4

        self.states_train: Optional[List[SkState]] = None
        self._surfaces: Optional[Dict[str, NDArray]] = None
        self.actions: List[SkAction] = [SkAction(x) for x in range(self.num_actions)]

        # check if data needs to be downloaded
        data_dir = get_data_dir()
        data_download_link: str = "https://github.com/forestagostinelli/DeepXubeData/raw/main/sokoban.tar.gz"
        if not os.path.exists(f"{data_dir}/sokoban/"):
            valid_user_in: bool = False
            while not valid_user_in:
                user_in: str = input(f"Sokoban data needs to be downloaded from {data_download_link}. "
                                     f"Download data (about 16MB)? (y/n):")
                if user_in.upper() == "Y":
                    valid_user_in = True
                    print("Downloading compressed data")
                    if not os.path.exists(data_dir):
                        os.makedirs(data_dir)

                    tar_gz_file_name = f"{data_dir}/sokoban.tar.gz"
                    wget.download(data_download_link, tar_gz_file_name, bar=None)
                    tar_gz_file = tarfile.open(tar_gz_file_name)
                    print("Uncompressing data")
                    tar_gz_file.extractall(data_dir)
                    print("Deleting compressed data")
                    os.remove(tar_gz_file_name)
                elif user_in.upper() == "N":
                    valid_user_in = True

    def next_state(self, states: List[SkState], actions: List[SkAction]) -> Tuple[List[SkState], List[float]]:
        agent = np.stack([state.agent for state in states], axis=0)
        boxes = np.stack([state.boxes for state in states], axis=0)
        walls_next = np.stack([state.walls for state in states], axis=0)

        idxs_arange = np.arange(0, len(states))
        agent_next_tmp = self._get_next_idx(agent, actions)
        agent_next = np.zeros(agent_next_tmp.shape, dtype=int)

        boxes_next = boxes.copy()

        # agent -> wall
        agent_wall = walls_next[idxs_arange, agent_next_tmp[:, 0], agent_next_tmp[:, 1]]
        agent_next[agent_wall] = agent[agent_wall]

        # agent -> box
        agent_box = boxes[idxs_arange, agent_next_tmp[:, 0], agent_next_tmp[:, 1]]
        boxes_next_tmp = self._get_next_idx(agent_next_tmp, actions)

        box_wall = walls_next[idxs_arange, boxes_next_tmp[:, 0], boxes_next_tmp[:, 1]]
        box_box = boxes[idxs_arange, boxes_next_tmp[:, 0], boxes_next_tmp[:, 1]]

        # agent -> box -> obstacle
        agent_box_obstacle = agent_box & (box_wall | box_box)
        agent_next[agent_box_obstacle] = agent[agent_box_obstacle]

        # agent -> box -> empty
        agent_box_empty = agent_box & np.logical_not(box_wall | box_box)
        agent_next[agent_box_empty] = agent_next_tmp[agent_box_empty]
        abe_idxs = np.where(agent_box_empty)[0]

        agent_next_idxs_abe = agent_next[agent_box_empty]
        boxes_next_idxs_abe = boxes_next_tmp[agent_box_empty]

        boxes_next[abe_idxs, agent_next_idxs_abe[:, 0], agent_next_idxs_abe[:, 1]] = False
        boxes_next[abe_idxs, boxes_next_idxs_abe[:, 0], boxes_next_idxs_abe[:, 1]] = True

        # agent -> empty
        agent_empty = np.logical_not(agent_wall | agent_box)
        agent_next[agent_empty] = agent_next_tmp[agent_empty]
        boxes_next[agent_empty] = boxes[agent_empty]

        states_next: List[SkState] = []
        for idx in range(len(states)):
            state_next: SkState = SkState(agent_next[idx], boxes_next[idx], walls_next[idx])
            states_next.append(state_next)

        transition_costs: List[float] = [1.0 for _ in range(len(states))]

        return states_next, transition_costs

    def get_actions_fixed(self) -> List[SkAction]:
        return self.actions.copy()

    def is_solved(self, states: List[SkState], goals: List[SkGoal]) -> List[bool]:
        boxes_states: NDArray = np.stack([state.boxes for state in states], axis=0)
        targets: NDArray = np.stack([goal.boxes for goal in goals], axis=0)
        return cast(List[bool], np.all(boxes_states == targets, axis=(1, 2)).tolist())

    def sample_start_states(self, num_states: int) -> List[SkState]:
        # get states
        if self.states_train is None:
            self.states_train = _get_train_states()
        state_idxs = np.random.randint(0, len(self.states_train), size=num_states)
        states: List[SkState] = [self.states_train[idx] for idx in state_idxs]

        # random walk
        step_range: Tuple[int, int] = (0, 100)

        steps_range: List[int] = list(range(step_range[0], step_range[1] + 1))
        step_nums: List[int] = np.random.choice(steps_range, num_states).tolist()

        return self.random_walk(states, step_nums)[0]

    def sample_goal_from_state(self, states_start: Optional[List[SkState]], states_goal: List[SkState]) -> List[SkGoal]:
        goals: List[SkGoal] = []
        for state_goal in states_goal:
            goals.append(SkGoal(state_goal.boxes))

        return goals

    def string_to_action(self, act_str: str) -> Optional[SkAction]:
        act_str_to_act: Dict[str, SkAction] = {"w": SkAction(0), "s": SkAction(1), "a": SkAction(2), "d": SkAction(3)}
        if act_str in act_str_to_act.keys():
            return act_str_to_act[act_str]
        else:
            return None

    def string_to_action_help(self) -> str:
        return "w, d, a, d: up, down, left, right"

    def visualize_state_goal(self, state: SkState, goal: SkGoal, fig: Figure) -> None:
        room_rgb = self.to_img(state, goal)

        ax = fig.add_subplot(111)
        ax.imshow(room_rgb)

    def to_img(self, state: SkState, goal: SkGoal) -> NDArray:
        if self._surfaces is None:
            self._surfaces = _get_surfaces()

        room_rgb: NDArray[np.uint8] = np.zeros(shape=(self.dim * 16, self.dim * 16, 3), dtype=np.uint8)
        for i in range(self.dim):
            x_i = i * 16

            for j in range(self.dim):
                y_j = j * 16

                surface_str: str
                if state.walls[i, j]:
                    surface_str = "wall"
                elif (state.agent[0] == i) and (state.agent[1] == j):
                    if goal.boxes[i, j]:
                        surface_str = "player_on_target"
                    else:
                        surface_str = "player"
                elif state.boxes[i, j]:
                    if goal.boxes[i, j]:
                        surface_str = "box_on_target"
                    else:
                        surface_str = "box"
                elif goal.boxes[i, j]:
                    surface_str = "box_target"
                else:
                    surface_str = "floor"
                room_rgb[x_i:(x_i + 16), y_j:(y_j + 16), :] = self._surfaces[surface_str]

        # img = Image.fromarray(room_rgb, 'RGB')
        # img = img.resize((self.img_dim, self.img_dim))

        return room_rgb

    def _get_next_idx(self, curr_idxs: NDArray[np.int_], actions: List[SkAction]) -> NDArray[np.int_]:
        actions_np: NDArray[np.int_] = np.array([action.action for action in actions])
        next_idxs: NDArray[np.int_] = curr_idxs.copy()

        action_idxs: NDArray[np.int_] = np.where(actions_np == 0)[0]
        next_idxs[action_idxs, 0] = next_idxs[action_idxs, 0] - 1

        action_idxs = np.where(actions_np == 1)[0]
        next_idxs[action_idxs, 0] = next_idxs[action_idxs, 0] + 1

        action_idxs = np.where(actions_np == 2)[0]
        next_idxs[action_idxs, 1] = next_idxs[action_idxs, 1] - 1

        action_idxs = np.where(actions_np == 3)[0]
        next_idxs[action_idxs, 1] = next_idxs[action_idxs, 1] + 1

        next_idxs = np.maximum(next_idxs, 0)
        next_idxs = np.minimum(next_idxs, self.dim - 1)

        return next_idxs

    def __getstate__(self) -> Dict:
        self.states_train = None
        self._surfaces = None

        return self.__dict__

    def __repr__(self) -> str:
        return "Sokoban"


@register_nnet_input("sokoban", "sokoban_nnet_input")
class SkNNetInput(FlatIn[Sokoban], StateGoalIn[Sokoban, SkState, SkGoal]):
    def get_input_info(self) -> Tuple[List[int], List[int]]:
        return [400], [1]

    def to_np(self, states: List[SkState], goals: List[SkGoal]) -> List[NDArray]:
        walls: NDArray = np.stack([state.walls for state in states], axis=0)
        boxes: NDArray = np.stack([state.boxes for state in states], axis=0)
        targets: NDArray = np.stack([goal.boxes for goal in goals], axis=0)
        agent_locs: NDArray = np.stack([state.agent for state in states], axis=0)
        agents: NDArray = np.zeros((len(states), self.domain.dim, self.domain.dim))
        agents[np.arange(0, len(states)), agent_locs[:, 0], agent_locs[:, 1]] = 1

        rep_np: NDArray = np.stack([walls, boxes, agents, targets], axis=1)
        rep_np = np.reshape(rep_np, (rep_np.shape[0], -1)).astype(np.uint8)
        return [rep_np]



from typing import Dict, Any, Generic, TypeVar, Type, Callable, Optional, List
from abc import ABC, abstractmethod
import logging


class Parser(ABC):
    @abstractmethod
    def parse(self, args_str: str) -> Dict[str, Any]:
        pass

    @abstractmethod
    def help(self) -> str:
        pass


T = TypeVar("T")


class Factory(Generic[T]):
    def __init__(self, class_type_str: str):
        self._class_registry: Dict[str, Type[T]] = dict()
        self._parser_registry: Dict[str, Type[Parser]] = dict()
        self._class_type_str: str = class_type_str

    def register_class(self, name: str) -> Callable[[Type[T]], Type[T]]:
        def deco(cls: Type[T]) -> Type[T]:
            if name in self._class_registry.keys():
                raise ValueError(f"{self._class_type_str.capitalize()} {name!r} already registered")
            self._class_registry[name] = cls
            return cls
        return deco

    def register_parser(self, name: str) -> Callable[[Type[Parser]], Type[Parser]]:
        def deco(cls: Type[Parser]) -> Type[Parser]:
            if name in self._parser_registry.keys():
                raise ValueError(f"{self._class_type_str.capitalize()} parser {name!r} already registered")
            self._parser_registry[name] = cls
            return cls
        return deco

    def get_parser(self, name: str) -> Optional[Parser]:
        cls_parser: Optional[Type[Parser]] = self._parser_registry.get(name)
        if cls_parser is not None:
            return cls_parser()
        else:
            return None

    def get_kwargs(self, name: str, args_str: Optional[str]) -> Dict[str, Any]:
        self.get_type(name)
        kwargs: Dict[str, Any] = dict()
        parser: Optional[Parser] = self.get_parser(name)
        if (parser is not None) and (args_str is not None):
            try:
                kwargs = parser.parse(args_str)
            except Exception as e:
                logging.exception(f"Error occurred: {e}")
                raise ValueError(f"Error parsing {args_str} for {self._class_type_str} {name!r}.\nParser help:\n{parser.help()}")
        else:
            assert args_str is None, f"No parser for {self._class_type_str} {name}, however, args given are {args_str}"
        return kwargs

    def get_type(self, name: str) -> Type[T]:
        try:
            return self._class_registry[name]
        except KeyError:
            raise ValueError(f"Unknown {self._class_type_str} {name!r}. Available: {sorted(self._class_registry)}")

    def build_class(self, name: str, kwargs: Dict[str, Any]) -> T:
        cls: Type[T] = self.get_type(name)
        return cls(**kwargs)

    def get_all_class_names(self) -> List[str]:
        return list(self._class_registry.keys())



from typing import Optional, List, Any
from dataclasses import dataclass

from deepxube.base.domain import Domain, State, Goal
from deepxube.base.heuristic import HeurNNet, PolicyNNet, HeurNNetPar, PolicyNNetPar, HeurFn, PolicyFn
from deepxube.base.pathfinding import PathFind, Instance
from deepxube.base.updater import UpdateHasHeur, UpdateHasPolicy, UpdateHeur, UpdatePolicy
from deepxube.base.trainer import TrainArgs
from deepxube.factories.pathfinding_factory import get_pathfind_functions
from deepxube.pathfinding.utils.performance import PathFindPerf
from deepxube.trainers.train_heur import TrainHeur
from deepxube.trainers.train_policy import TrainPolicy
from deepxube.utils.command_line_utils import get_pathfind_from_arg, get_pathfind_name_kwargs
from deepxube.nnet import nnet_utils
from deepxube.utils import data_utils

from torch.utils.tensorboard import SummaryWriter

import sys
import os

import torch
import time


@dataclass
class TestArgs:
    test_states: List[State]
    test_goals: List[Goal]
    search_itrs: int
    pathfinds: List[str]
    test_nnet_batch_size: int
    test_up_freq: int
    test_init: bool

    def __repr__(self) -> str:
        return (f"TestArgs(instances={len(self.test_states)}, search_itrs={self.search_itrs}, "
                f"pathfinds={self.pathfinds}, test_nnet_batch_size={self.test_nnet_batch_size}, "
                f"test_up_freq={self.test_up_freq}, test_init={self.test_init})")


def get_curr_itr(train_heur: Optional[TrainHeur], train_policy: Optional[TrainPolicy]) -> int:
    if train_heur is not None:
        return train_heur.status.itr
    else:
        assert train_policy is not None
        return train_policy.status.itr


def get_curr_update_num(train_heur: Optional[TrainHeur], train_policy: Optional[TrainPolicy]) -> int:
    if train_policy is not None:
        return train_policy.status.update_num
    else:
        assert train_heur is not None
        return train_heur.status.update_num


def get_pathfind(domain: Domain, heur_nnet_par: Optional[HeurNNetPar], train_heur: Optional[TrainHeur], policy_nnet_par: Optional[PolicyNNetPar],
                 train_policy: Optional[TrainPolicy], test_nnet_batch_size: int, pathfind_arg: str) -> PathFind:
    heur_fn: Optional[HeurFn] = None
    policy_fn: Optional[PolicyFn] = None
    if heur_nnet_par is not None:
        assert train_heur is not None
        heur_fn = heur_nnet_par.get_nnet_fn(train_heur.nnet, test_nnet_batch_size, train_heur.device, None)
    if policy_nnet_par is not None:
        assert train_policy is not None
        policy_fn = policy_nnet_par.get_nnet_fn(train_policy.nnet, test_nnet_batch_size, train_policy.device, None)

    pathfind_functions: Any = get_pathfind_functions(get_pathfind_name_kwargs(pathfind_arg)[0], heur_fn, policy_fn)
    pathfind: PathFind = get_pathfind_from_arg(domain, pathfind_functions, pathfind_arg)[0]

    return pathfind


def train(domain: Domain, heur_nnet_par: Optional[HeurNNetPar], update_heur: Optional[UpdateHeur], policy_nnet_par: Optional[PolicyNNetPar],
          update_policy: Optional[UpdatePolicy], nnet_dir: str, train_args: TrainArgs, test_args: Optional[TestArgs] = None,
          debug: bool = False) -> None:
    if not os.path.exists(nnet_dir):
        os.makedirs(nnet_dir)

    # logging
    if not debug:
        output_save_loc = f"{nnet_dir}/output.txt"
        sys.stdout = data_utils.Logger(output_save_loc, "a")

    if 'SLURM_JOB_ID' in os.environ:
        print("SLURM JOB ID: %s" % os.environ['SLURM_JOB_ID'])

    # get device
    on_gpu: bool
    device: torch.device
    device, devices, on_gpu = nnet_utils.get_device()
    print("device: %s, devices: %s, on_gpu: %s" % (device, devices, on_gpu))

    # file info
    writer: SummaryWriter = SummaryWriter(nnet_dir)

    heur_file: str = f"{nnet_dir}/heur.pt"
    heur_targ_file: str = f"{nnet_dir}/heur_targ.pt"
    heur_status_file: str = f"{nnet_dir}/heur_status.pkl"
    heur_train_summ_file: str = f"{nnet_dir}/heur_train_summary.pkl"

    policy_file = f"{nnet_dir}/policy.pt"
    policy_targ_file = f"{nnet_dir}/policy_targ.pt"
    policy_status_file: str = f"{nnet_dir}/policy_status.pkl"
    policy_train_summ_file: str = f"{nnet_dir}/policy_train_summary.pkl"

    # set updater heur info
    if heur_nnet_par is not None:
        for updater in [update_heur, update_policy]:
            if (updater is not None) and isinstance(updater, UpdateHasHeur):
                updater.set_heur_nnet(heur_nnet_par)
                updater.set_heur_file(heur_targ_file)
    if policy_nnet_par is not None:
        for updater in [update_heur, update_policy]:
            if (updater is not None) and isinstance(updater, UpdateHasPolicy):
                updater.set_policy_nnet(policy_nnet_par)
                updater.set_policy_file(policy_targ_file)

    # start_procs and trainers
    train_heur: Optional[TrainHeur] = None
    train_policy: Optional[TrainPolicy] = None
    if heur_nnet_par is not None:
        assert update_heur is not None

        heur_nnet: HeurNNet = heur_nnet_par.get_nnet()
        print(heur_nnet_par)
        to_main_q, from_main_qs = update_heur.start_procs(train_args.rb * train_args.batch_size * update_heur.up_args.up_itrs)
        train_heur = TrainHeur(heur_nnet, update_heur, to_main_q, from_main_qs, heur_file, heur_targ_file, heur_status_file, heur_train_summ_file, device,
                               on_gpu, writer, train_args)
    if policy_nnet_par is not None:
        assert update_policy is not None

        policy_nnet: PolicyNNet = policy_nnet_par.get_nnet()
        print(policy_nnet_par)
        to_main_q, from_main_qs = update_policy.start_procs(train_args.rb * train_args.batch_size * update_policy.up_args.up_itrs)
        train_policy = TrainPolicy(policy_nnet, update_policy, to_main_q, from_main_qs, policy_file, policy_targ_file, policy_status_file,
                                   policy_train_summ_file, device, on_gpu, writer, train_args)

    # print info
    for updater in [update_heur, update_policy]:
        if updater is not None:
            print(f"{updater}")
            print(get_pathfind(domain, heur_nnet_par, train_heur, policy_nnet_par, train_policy, 100, updater.pathfind_arg))

    print(f"{train_args}")
    print(domain)
    if test_args is not None:
        print(f"{test_args}")

    # training
    up_itr_performed: bool = False

    curr_itr: int = get_curr_itr(train_heur, train_policy)
    while curr_itr < train_args.max_itrs:
        # test
        do_test: bool = False
        if test_args is not None:
            update_num: int = get_curr_update_num(train_heur, train_policy)
            if update_num > 0:
                do_test = update_num % test_args.test_up_freq == 0
            elif update_num == 0:
                do_test = test_args.test_init

        if do_test:
            assert test_args is not None
            test(domain, heur_nnet_par, train_heur, policy_nnet_par, train_policy, test_args, writer, curr_itr)

        # train
        for train_obj in [train_heur, train_policy]:
            if isinstance(train_obj, TrainHeur) and train_args.skip_heur:
                continue
            if isinstance(train_obj, TrainPolicy) and train_args.skip_policy:
                continue

            for updater in [update_heur, update_policy]:
                if updater is None:
                    continue
                if train_heur is not None:
                    assert isinstance(updater, UpdateHasHeur)
                    updater.set_targ_update_num(updater.heur_name(), train_heur.status.targ_update_num)
                if train_policy is not None:
                    assert isinstance(updater, UpdateHasPolicy)
                    updater.set_targ_update_num(updater.policy_name(), train_policy.status.targ_update_num)

            if (train_obj is None) or (train_obj.status.itr > curr_itr):
                continue
            train_obj.update_step()
            torch.cuda.empty_cache()

        up_itr_performed = True
        curr_itr = get_curr_itr(train_heur, train_policy)

    # test
    if (test_args is not None) and up_itr_performed:
        test(domain, heur_nnet_par, train_heur, policy_nnet_par, train_policy, test_args, writer, curr_itr)

    # stop procs
    for updater in [update_heur, update_policy]:
        if updater is not None:
            updater.stop_procs()

    print("Done")


def get_pathfind_w_instances(domain: Domain, heur_nnet_par: Optional[HeurNNetPar], train_heur: Optional[TrainHeur], policy_nnet_par: Optional[PolicyNNetPar],
                             train_policy: Optional[TrainPolicy], test_args: TestArgs, pathfind_arg: str) -> PathFind:
    pathfind: PathFind = get_pathfind(domain, heur_nnet_par, train_heur, policy_nnet_par, train_policy, test_args.test_nnet_batch_size, pathfind_arg)

    instances: List[Instance] = pathfind.make_instances(test_args.test_states, test_args.test_goals, None, True)
    pathfind.add_instances(instances)

    return pathfind


def test(domain: Domain, heur_nnet_par: Optional[HeurNNetPar], train_heur: Optional[TrainHeur], policy_nnet_par: Optional[PolicyNNetPar],
         train_policy: Optional[TrainPolicy], test_args: TestArgs, writer: SummaryWriter, curr_itr: int) -> None:
    print(f"Testing - itr: {curr_itr}, num_inst: {len(test_args.test_states)}, num_pathfinds: {len(test_args.pathfinds)}")
    for pathfind_idx in range(len(test_args.pathfinds)):
        start_time = time.time()
        # get pathfinding alg with test instances
        pathfind_arg: str = test_args.pathfinds[pathfind_idx]
        pathfind: PathFind = get_pathfind_w_instances(domain, heur_nnet_par, train_heur, policy_nnet_par, train_policy, test_args, pathfind_arg)

        # attempt to solve
        for _ in range(test_args.search_itrs):
            pathfind.step()

        # get performacne
        pathfind_perf: PathFindPerf = PathFindPerf()
        for instance in pathfind.instances:
            pathfind_perf.update_perf(instance)
        test_time = time.time() - start_time

        # log
        per_solved_ave, path_cost_ave, search_itrs_ave = pathfind_perf.stats()
        test_info_l: List[str] = [f"%solved: {per_solved_ave:.2f}", f"path_costs: {path_cost_ave:.3f}",
                                  f"search_itrs: {search_itrs_ave:.3f}",
                                  f"test_time: {test_time:.2f}"]
        writer.add_scalar(f"val/{pathfind_arg}/solved", per_solved_ave, curr_itr)
        writer.add_scalar(f"val/{pathfind_arg}/path_cost", path_cost_ave, curr_itr)
        writer.add_scalar(f"val/{pathfind_arg}/search_itrs", search_itrs_ave, curr_itr)
        print(f"Test {pathfind_arg} - {', '.join(test_info_l)}")



from typing import List, Dict, Tuple, Optional, cast
from deepxube.base.domain import State, Action, Goal, GoalStartRevWalkableActsRev, NextStateNPActsEnumFixed, StateGoalVizable, StringToAct
from deepxube.base.nnet_input import HasFlatSGActsEnumFixedIn, HasFlatSGAIn
from deepxube.factories.domain_factory import domain_factory
from matplotlib import pyplot as plt
from matplotlib.figure import Figure


import numpy as np

from numpy.typing import NDArray


class Cube3State(State):
    __slots__ = ['colors', 'hash']

    def __init__(self, colors: NDArray[np.uint8]) -> None:
        self.colors: NDArray[np.uint8] = colors
        self.hash: Optional[int] = None

    def __hash__(self) -> int:
        if self.hash is None:
            self.hash = hash(self.colors.tobytes())
        return self.hash

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Cube3State):
            return np.array_equal(self.colors, other.colors)
        return NotImplemented


class Cube3Goal(Goal):
    def __init__(self, colors: NDArray[np.uint8]):
        self.colors: NDArray[np.uint8] = colors


class Cube3Action(Action):
    def __init__(self, action: int) -> None:
        self.action = action

    def __hash__(self) -> int:
        return self.action

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Cube3Action):
            return self.action == other.action
        return NotImplemented


def _get_adj() -> Dict[int, NDArray[np.int_]]:
    # WHITE:0, YELLOW:1, BLUE:2, GREEN:3, ORANGE: 4, RED: 5
    return {0: np.array([2, 5, 3, 4]),
            1: np.array([2, 4, 3, 5]),
            2: np.array([0, 4, 1, 5]),
            3: np.array([0, 5, 1, 4]),
            4: np.array([0, 3, 1, 2]),
            5: np.array([0, 2, 1, 3])
            }


class Quaternion:
    """Quaternion Rotation:

    Class to aid in representing 3D rotations via quaternions.
    """

    @classmethod
    def from_v_theta(cls, v, theta) -> 'Quaternion':  # type: ignore
        """
        Construct quaternions from unit vectors v and rotation angles theta

        Parameters
        ----------
        v : array_like
            array of vectors, last dimension 3. Vectors will be normalized.
        theta : array_like
            array of rotation angles in radians, shape = v.shape[:-1].

        Returns
        -------
        q : quaternion object
            quaternion representing the rotations
        """
        theta = np.asarray(theta)
        v = np.asarray(v)
        s = np.sin(0.5 * theta)
        c = np.cos(0.5 * theta)

        v = v * s / np.sqrt(np.sum(v * v, -1))
        x_shape = v.shape[:-1] + (4,)

        x: NDArray = np.ones(x_shape).reshape(-1, 4)
        x[:, 0] = c.ravel()
        x[:, 1:] = v.reshape(-1, 3)
        x = x.reshape(x_shape)

        return cls(x)

    def __init__(self, x: NDArray):
        self.x = np.asarray(x, dtype=float)

    def __repr__(self) -> str:
        return "Quaternion:\n" + self.x.__repr__()

    def __mul__(self, other: 'Quaternion') -> 'Quaternion':
        # multiplication of two quaternions.
        # we don't implement multiplication by a scalar
        sxr = self.x.reshape(self.x.shape[:-1] + (4, 1))
        oxr = other.x.reshape(other.x.shape[:-1] + (1, 4))

        prod = sxr * oxr
        return_shape = prod.shape[:-1]
        prod = prod.reshape((-1, 4, 4)).transpose((1, 2, 0))

        ret = np.array([(prod[0, 0] - prod[1, 1]
                         - prod[2, 2] - prod[3, 3]),
                        (prod[0, 1] + prod[1, 0]
                         + prod[2, 3] - prod[3, 2]),
                        (prod[0, 2] - prod[1, 3]
                         + prod[2, 0] + prod[3, 1]),
                        (prod[0, 3] + prod[1, 2]
                         - prod[2, 1] + prod[3, 0])],
                       dtype=float,
                       order='F').T
        return self.__class__(ret.reshape(return_shape))

    def as_v_theta(self) -> Tuple[NDArray, NDArray]:
        """Return the v, theta equivalent of the (normalized) quaternion"""
        x = self.x.reshape((-1, 4)).T

        # compute theta
        norm = np.sqrt((x ** 2).sum(0))
        theta = 2 * np.arccos(x[0] / norm)

        # compute the unit vector
        v = np.array(x[1:], order='F', copy=True)
        v /= np.sqrt(np.sum(v ** 2, 0))

        # reshape the results
        v = v.T.reshape(self.x.shape[:-1] + (3,))
        theta = theta.reshape(self.x.shape[:-1])

        return v, theta

    def as_rotation_matrix(self) -> NDArray:
        """Return the rotation matrix of the (normalized) quaternion"""
        v, theta = self.as_v_theta()

        shape = theta.shape
        theta = theta.reshape(-1)
        v = v.reshape(-1, 3).T
        c = np.cos(theta)
        s = np.sin(theta)

        mat = np.array([[v[0] * v[0] * (1. - c) + c,
                         v[0] * v[1] * (1. - c) - v[2] * s,
                         v[0] * v[2] * (1. - c) + v[1] * s],
                        [v[1] * v[0] * (1. - c) + v[2] * s,
                         v[1] * v[1] * (1. - c) + c,
                         v[1] * v[2] * (1. - c) - v[0] * s],
                        [v[2] * v[0] * (1. - c) - v[1] * s,
                         v[2] * v[1] * (1. - c) + v[0] * s,
                         v[2] * v[2] * (1. - c) + c]],
                       order='F').T
        return mat.reshape(shape + (3, 3))

    def rotate(self, points):  # type: ignore
        rot_mat = self.as_rotation_matrix()
        return np.dot(points, rot_mat.T)


def project_points(points, q: Quaternion, view, vertical) -> NDArray:  # type: ignore
    """Project points using a quaternion q and a view v

    Parameters
    ----------
    points : array_like
        array of last-dimension 3
    q : utils.viz_utils.Quaternion
        quaternion representation of the rotation
    view : array_like
        length-3 vector giving the point of view
    vertical : array_like
        direction of y-axis for view.  An error will be raised if it
        is parallel to the view.

    Returns
    -------
    proj: array_like
        array of projected points: same shape as points.
    """
    if vertical is None:
        vertical = [0, 1, 0]
    points = np.asarray(points)
    view = np.asarray(view)

    xdir = np.cross(vertical, view).astype(float)

    if np.all(xdir == 0):
        raise ValueError("vertical is parallel to v")

    xdir /= np.sqrt(np.dot(xdir, xdir))

    # get the unit vector corresponing to vertical
    ydir = np.cross(view, xdir)
    ydir /= np.sqrt(np.dot(ydir, ydir))

    # normalize the viewer location: this is the z-axis
    v2 = np.dot(view, view)
    zdir = view / np.sqrt(v2)

    # rotate the points
    rot_mat = q.as_rotation_matrix()
    r_pts = np.dot(points, rot_mat.T)

    # project the points onto the view
    dpoint = r_pts - view
    dpoint_view = np.dot(dpoint, view).reshape(dpoint.shape[:-1] + (1,))
    dproj = -dpoint * v2 / dpoint_view

    trans = list(range(1, dproj.ndim)) + [0]
    return np.array([np.dot(dproj, xdir),
                     np.dot(dproj, ydir),
                     -np.dot(dpoint, zdir)]).transpose(trans)


class InteractiveCube(plt.Axes):  # type: ignore
    # Define some attributes
    base_face = np.array([[1, 1, 1],
                          [1, -1, 1],
                          [-1, -1, 1],
                          [-1, 1, 1],
                          [1, 1, 1]], dtype=float)
    stickerwidth = 0.9
    stickermargin = 0.5 * (1. - stickerwidth)
    stickerthickness = 0.001
    (d1, d2, d3) = (1 - stickermargin,
                    1 - 2 * stickermargin,
                    1 + stickerthickness)
    base_sticker = np.array([[d1, d2, d3], [d2, d1, d3],
                             [-d2, d1, d3], [-d1, d2, d3],
                             [-d1, -d2, d3], [-d2, -d1, d3],
                             [d2, -d1, d3], [d1, -d2, d3],
                             [d1, d2, d3]], dtype=float)

    base_face_centroid = np.array([[0, 0, 1]])
    base_sticker_centroid = np.array([[0, 0, 1 + stickerthickness]])

    def __init__(self, n, colors: NDArray, view=(0, 0, 10), fig=None, **kwargs) -> None:  # type: ignore
        self.colors: NDArray = colors

        # Define rotation angles and axes for the six sides of the cube
        x, y, z = np.eye(3)
        self.rots = [Quaternion.from_v_theta(x, np.asarray(theta)) for theta in (np.pi / 2, -np.pi / 2)]
        self.rots += [Quaternion.from_v_theta(y, np.asarray(theta)) for theta in (np.pi / 2, -np.pi / 2, np.pi, 2 * np.pi)]

        rect = [0, 0.16, 1, 0.84]
        self._move_list: List = []

        self.N = n
        self._prevStates: List = []

        self._view = view
        self._start_rot: Quaternion = Quaternion.from_v_theta(np.asarray((1, -1, 0)), np.asarray(-np.pi / 6))

        self._grey_stickers: List = []
        self._black_stickers: List = []

        if fig is None:
            fig = plt.gcf()

        # disable default key press events
        # callbacks = fig.canvas.callbacks.callbacks
        # del callbacks['key_press_event']

        # add some defaults, and draw axes
        kwargs.update(dict(aspect=kwargs.get('aspect', 'equal'),
                           xlim=kwargs.get('xlim', (-1.7, 1.5)),
                           ylim=kwargs.get('ylim', (-1.5, 1.7)),
                           frameon=kwargs.get('frameon', False),
                           xticks=kwargs.get('xticks', []),
                           yticks=kwargs.get('yticks', [])))
        super(InteractiveCube, self).__init__(fig, rect, **kwargs)
        self.xaxis.set_major_formatter(plt.NullFormatter())  # type: ignore
        self.yaxis.set_major_formatter(plt.NullFormatter())  # type: ignore

        self._start_xlim = kwargs['xlim']
        self._start_ylim = kwargs['ylim']

        self._active = False  # true when mouse is over axes
        self._button1 = False  # true when button 1 is pressed
        self._button2 = False  # true when button 2 is pressed
        self._tab = False  # tab key pressed

        # Define movement for up/down arrows or up/down mouse movement
        self._ax_UD = (1, 0, 0)
        self._step_UD = 0.01

        # Define movement for left/right arrows or left/right mouse movement
        self._ax_LR = (0, -1, 0)
        self._step_LR = 0.01

        self._ax_LR_alt = (0, 0, 1)

        self._current_rot: Quaternion = self._start_rot  # current rotation state
        self._face_polys: Optional[List] = None
        self._sticker_polys: Optional[List] = None

        self.plastic_color = 'black'

        # WHITE:0 - U, YELLOW:1 - D, BLUE:2 - L, GREEN:3 - R, ORANGE: 4 - B, RED: 5 - F
        self.face_colors = ["w", "#ffcf00",
                            "#ff6f00", "#cf0000",
                            "#00008f", "#009f0f",
                            "gray", "none"]

        self._initialize_arrays()

        self.figure.canvas.mpl_connect('button_press_event',
                                       self._mouse_press)
        self.figure.canvas.mpl_connect('button_release_event',
                                       self._mouse_release)
        self.figure.canvas.mpl_connect('motion_notify_event',
                                       self._mouse_motion)

        self._draw_cube()
        # self._initialize_widgets()

    def set_rot(self, rot: int) -> None:
        if rot == 0:
            self._current_rot = Quaternion.from_v_theta(np.asarray((-0.53180525, 0.83020462, 0.16716299)), np.asarray(0.95063829))
        elif rot == 1:
            self._current_rot = Quaternion.from_v_theta(np.asarray((0.9248325, 0.14011997, -0.35362584)), np.asarray(2.49351394))

        self._draw_cube()

    def _initialize_arrays(self) -> None:
        # initialize centroids, faces, and stickers.  We start with a
        # base for each one, and then translate & rotate them into position.

        # Define N^2 translations for each face of the cube
        cubie_width = 2. / self.N
        translations = np.array([[[-1 + (i + 0.5) * cubie_width,
                                   -1 + (j + 0.5) * cubie_width, 0]]
                                 for i in range(self.N)
                                 for j in range(self.N)])

        # Create arrays for centroids, faces, stickers
        face_centroids = []
        faces = []
        sticker_centroids = []
        stickers = []
        colors = []

        factor = np.array([1. / self.N, 1. / self.N, 1])

        for i in range(6):
            rot_mat = self.rots[i].as_rotation_matrix()
            faces_t = np.dot(factor * self.base_face
                             + translations, rot_mat.T)
            stickers_t = np.dot(factor * self.base_sticker
                                + translations, rot_mat.T)
            face_centroids_t = np.dot(self.base_face_centroid
                                      + translations, rot_mat.T)
            sticker_centroids_t = np.dot(self.base_sticker_centroid
                                         + translations, rot_mat.T)
            # colors_i = i + np.zeros(face_centroids_t.shape[0], dtype=int)
            colors_i = np.arange(i * face_centroids_t.shape[0], (i + 1) * face_centroids_t.shape[0])

            # append face ID to the face centroids for lex-sorting
            face_centroids_t = np.hstack([face_centroids_t.reshape(-1, 3),
                                          colors_i[:, None]])
            sticker_centroids_t = sticker_centroids_t.reshape((-1, 3))

            faces.append(faces_t)
            face_centroids.append(face_centroids_t)
            stickers.append(stickers_t)
            sticker_centroids.append(sticker_centroids_t)

            colors.append(colors_i)

        self._face_centroids = np.vstack(face_centroids)
        self._faces = np.vstack(faces)
        self._sticker_centroids = np.vstack(sticker_centroids)
        self._stickers = np.vstack(stickers)

    def rotate(self, rot) -> None:  # type: ignore
        self._current_rot = self._current_rot * rot

    def _project(self, pts):  # type: ignore
        return project_points(pts, self._current_rot, self._view, [0, 1, 0])

    def _draw_cube(self) -> None:
        stickers = self._project(self._stickers)[:, :, :2]
        faces = self._project(self._faces)[:, :, :2]
        face_centroids = self._project(self._face_centroids[:, :3])
        sticker_centroids = self._project(self._sticker_centroids[:, :3])

        plastic_color = self.plastic_color
        # self._colors[np.ravel_multi_index((0,1,2),(6,N,N))] = 10
        colors = np.asarray(self.face_colors)[self.colors]
        for idx in self._grey_stickers:
            colors[idx] = "grey"
        for idx in self._black_stickers:
            colors[idx] = "k"

        face_zorders = -face_centroids[:, 2]
        sticker_zorders = -sticker_centroids[:, 2]

        if self._face_polys is None:
            # initial call: create polygon objects and add to axes
            self._face_polys = []
            self._sticker_polys = []

            for i in range(len(colors)):
                fp = plt.Polygon(faces[i], facecolor=plastic_color, zorder=face_zorders[i])  # type: ignore
                sp = plt.Polygon(stickers[i], facecolor=colors[i], zorder=sticker_zorders[i])  # type: ignore

                self._face_polys.append(fp)
                self._sticker_polys.append(sp)
                self.add_patch(fp)
                self.add_patch(sp)
        else:
            assert self._sticker_polys is not None
            # subsequent call: updater the polygon objects
            for i in range(len(colors)):
                self._face_polys[i].set_xy(faces[i])
                self._face_polys[i].set_zorder(face_zorders[i])
                self._face_polys[i].set_facecolor(plastic_color)

                self._sticker_polys[i].set_xy(stickers[i])
                self._sticker_polys[i].set_zorder(sticker_zorders[i])
                self._sticker_polys[i].set_facecolor(colors[i])

        self.figure.canvas.draw()

    def _mouse_press(self, event, event_x=None, event_y=None):  # type: ignore
        if event_x is not None and event_y is not None:
            self._event_xy = (event_x, event_y)
            self._button1 = True
        else:
            self._event_xy = (event.x, event.y)
            if event.button == 1:
                self._button1 = True
            elif event.button == 3:
                self._button2 = True

    def _mouse_release(self, event):  # type: ignore
        self._event_xy = None  # type: ignore
        if event.button == 1:
            self._button1 = False
        elif event.button == 3:
            self._button2 = False

    def _mouse_motion(self, event, event_x=None, event_y=None):  # type: ignore
        if self._button1 or self._button2:
            if event_x is not None and event_y is not None:
                dx = event_x - self._event_xy[0]
                dy = event_y - self._event_xy[1]
                self._event_xy = (event_x, event_y)
            else:
                dx = event.x - self._event_xy[0]
                dy = event.y - self._event_xy[1]
                self._event_xy = (event.x, event.y)

            if self._button1:
                if self._tab:
                    ax_lr = self._ax_LR_alt
                else:
                    ax_lr = self._ax_LR
                rot1 = Quaternion.from_v_theta(self._ax_UD, self._step_UD * dy)
                rot2 = Quaternion.from_v_theta(ax_lr, self._step_LR * dx)
                self.rotate(rot1 * rot2)

                self._draw_cube()

            if self._button2:
                factor = 1 - 0.003 * (dx + dy)
                xlim = self.get_xlim()
                ylim = self.get_ylim()
                self.set_xlim(factor * xlim[0], factor * xlim[1])
                self.set_ylim(factor * ylim[0], factor * ylim[1])

                self.figure.canvas.draw()


@domain_factory.register_class("cube3")
class Cube3(NextStateNPActsEnumFixed[Cube3State, Cube3Action, Cube3Goal],
            GoalStartRevWalkableActsRev[Cube3State, Cube3Action, Cube3Goal],
            HasFlatSGActsEnumFixedIn[Cube3State, Cube3Action, Cube3Goal], HasFlatSGAIn[Cube3State, Cube3Action, Cube3Goal],
            StateGoalVizable[Cube3State, Cube3Action, Cube3Goal], StringToAct[Cube3State, Cube3Action, Cube3Goal]):
    atomic_actions: List[str] = ["%s%i" % (f, n) for f in ['U', 'D', 'L', 'R', 'B', 'F'] for n in [-1, 1]]

    def __init__(self) -> None:
        super().__init__()
        self.cube_len: int = 3
        self.num_colors: int = 6
        self.num_actions = len(self.atomic_actions)
        self.num_stickers: int = self.num_colors * (self.cube_len ** 2)

        # solved state
        self.goal_colors: NDArray[np.uint8] = (np.arange(0, self.num_stickers, 1,
                                                         dtype=np.uint8) // (self.cube_len ** 2)).astype(np.uint8)

        # get idxs changed for moves
        self.rotate_idxs_new: Dict[str, NDArray[np.int_]]
        self.rotate_idxs_old: Dict[str, NDArray[np.int_]]

        self.adj_faces: Dict[int, NDArray[np.int_]] = _get_adj()

        self.rotate_idxs_new, self.rotate_idxs_old = self._compute_rotation_idxs(self.cube_len, self.atomic_actions)
        self.actions: List[Cube3Action] = [Cube3Action(x) for x in range(self.num_actions)]

    def is_solved(self, states: List[Cube3State], goals: List[Cube3Goal]) -> List[bool]:
        states_np: NDArray = np.stack([x.colors for x in states], axis=0)
        goals_np: NDArray = np.stack([x.colors for x in goals], axis=0)
        return cast(List[bool], np.all(states_np == goals_np, axis=1).tolist())

    def get_goal_states(self, num_states: int) -> List[Cube3State]:
        return [Cube3State(self.goal_colors.copy()) for _ in range(num_states)]

    def sample_goalstate_goal_pairs(self, num: int) -> Tuple[List[Cube3State], List[Cube3Goal]]:
        states_goal: List[Cube3State] = [Cube3State(self.goal_colors.copy())] * num
        goals: List[Cube3Goal] = [Cube3Goal(self.goal_colors.copy())] * num

        return states_goal, goals

    def get_input_info_flat_sg(self) -> Tuple[List[int], List[int]]:
        return [self.num_stickers], [self.num_colors]

    def get_input_info_flat_sga(self) -> Tuple[List[int], List[int]]:
        return [self.num_stickers, 1], [self.num_colors, self.get_num_acts()]

    def to_np_flat_sg(self, states: List[Cube3State], goals: List[Cube3Goal]) -> List[NDArray]:
        return [np.stack([x.colors for x in states], axis=0).astype(np.uint8)]

    def to_np_flat_sga(self, states: List[Cube3State], goals: List[Cube3Goal],
                       actions: List[Cube3Action]) -> List[NDArray]:
        return self.to_np_flat_sg(states, goals) + [np.expand_dims(np.array(self.actions_to_indices(actions)), 1)]

    def actions_to_indices(self, actions: List[Cube3Action]) -> List[int]:
        return [action_cube3.action for action_cube3 in actions]

    def visualize_state_goal(self, state: Cube3State, goal: Cube3Goal, fig: Figure) -> None:
        interactive_cube: InteractiveCube = InteractiveCube(3, state.colors)
        fig.add_axes(interactive_cube)

    def string_to_action(self, act_str: str) -> Optional[Cube3Action]:
        if act_str in self.atomic_actions:
            return Cube3Action(self.atomic_actions.index(act_str))
        else:
            return None

    def string_to_action_help(self) -> str:
        return "<face><dir>. i.e. U1, U-1, faces are U, D, L, R, B, F and dirs are 1 and -1"

    def get_actions_fixed(self) -> List[Cube3Action]:
        return self.actions.copy()

    def rev_action(self, states: List[Cube3State], actions: List[Cube3Action]) -> List[Cube3Action]:
        actions_rev: List[Cube3Action] = []
        for action in actions:
            action_val: int = action.action
            action_val_rev: int
            if action_val % 2 == 0:
                action_val_rev = action_val + 1
            else:
                action_val_rev = action_val - 1
            actions_rev.append(Cube3Action(action_val_rev))

        return actions_rev

    def _states_to_np(self, states: List[Cube3State]) -> List[NDArray[np.uint8]]:
        return [np.stack([x.colors for x in states], axis=0)]

    def _np_to_states(self, states_np: List[NDArray]) -> List[Cube3State]:
        assert len(states_np) == 1
        return [Cube3State(x) for x in states_np[0]]

    def _next_state_np(self, states_np_l: List[NDArray[np.uint8]],
                       actions: List[Cube3Action]) -> Tuple[List[NDArray], List[float]]:
        assert len(states_np_l) == 1
        colors_next_np: NDArray[np.uint8] = states_np_l[0].copy()
        assert colors_next_np.shape[0] == len(actions), f"#states {colors_next_np.shape[0]} != #actions {len(actions)}"

        state_idxs: NDArray = np.arange(0, colors_next_np.shape[0])
        state_idxs = np.expand_dims(state_idxs, 1)

        rotate_idxs_new: NDArray = np.stack([self.rotate_idxs_new[self.atomic_actions[action.action]] for action in actions])
        rotate_idxs_old: NDArray = np.stack([self.rotate_idxs_old[self.atomic_actions[action.action]] for action in actions])
        colors_next_np[state_idxs, rotate_idxs_new] = colors_next_np[state_idxs, rotate_idxs_old]

        return [colors_next_np], [1.0] * len(actions)

    def _compute_rotation_idxs(self, cube_len: int,
                               moves: List[str]) -> Tuple[Dict[str, NDArray[np.int_]], Dict[str, NDArray[np.int_]]]:
        rotate_idxs_new: Dict[str, NDArray[np.int_]] = dict()
        rotate_idxs_old: Dict[str, NDArray[np.int_]] = dict()

        for move in moves:
            f: str = move[0]
            sign: int = int(move[1:])

            rotate_idxs_new[move] = np.array([], dtype=int)
            rotate_idxs_old[move] = np.array([], dtype=int)

            colors: NDArray = np.zeros((6, cube_len, cube_len), dtype=np.int64)
            colors_new = np.copy(colors)

            # WHITE:0, YELLOW:1, BLUE:2, GREEN:3, ORANGE: 4, RED: 5

            adj_idxs = {0: {2: [range(0, cube_len), cube_len - 1], 3: [range(0, cube_len), cube_len - 1],
                            4: [range(0, cube_len), cube_len - 1], 5: [range(0, cube_len), cube_len - 1]},
                        1: {2: [range(0, cube_len), 0], 3: [range(0, cube_len), 0], 4: [range(0, cube_len), 0],
                            5: [range(0, cube_len), 0]},
                        2: {0: [0, range(0, cube_len)], 1: [0, range(0, cube_len)],
                            4: [cube_len - 1, range(cube_len - 1, -1, -1)], 5: [0, range(0, cube_len)]},
                        3: {0: [cube_len - 1, range(0, cube_len)], 1: [cube_len - 1, range(0, cube_len)],
                            4: [0, range(cube_len - 1, -1, -1)], 5: [cube_len - 1, range(0, cube_len)]},
                        4: {0: [range(0, cube_len), cube_len - 1], 1: [range(cube_len - 1, -1, -1), 0],
                            2: [0, range(0, cube_len)], 3: [cube_len - 1, range(cube_len - 1, -1, -1)]},
                        5: {0: [range(0, cube_len), 0], 1: [range(cube_len - 1, -1, -1), cube_len - 1],
                            2: [cube_len - 1, range(0, cube_len)], 3: [0, range(cube_len - 1, -1, -1)]}
                        }
            face_dict = {'U': 0, 'D': 1, 'L': 2, 'R': 3, 'B': 4, 'F': 5}
            face = face_dict[f]

            faces_to = self.adj_faces[face]
            if sign == 1:
                faces_from = faces_to[(np.arange(0, len(faces_to)) + 1) % len(faces_to)]
            else:
                faces_from = faces_to[(np.arange(len(faces_to) - 1, len(faces_to) - 1 + len(faces_to))) % len(faces_to)]

            cubes_idxs = [[0, range(0, cube_len)], [range(0, cube_len), cube_len - 1],
                          [cube_len - 1, range(cube_len - 1, -1, -1)], [range(cube_len - 1, -1, -1), 0]]
            cubes_to = np.array([0, 1, 2, 3])
            if sign == 1:
                cubes_from = cubes_to[(np.arange(len(cubes_to) - 1, len(cubes_to) - 1 + len(cubes_to))) % len(cubes_to)]
            else:
                cubes_from = cubes_to[(np.arange(0, len(cubes_to)) + 1) % len(cubes_to)]

            for i in range(4):
                idxs_new = [[idx1, idx2] for idx1 in np.array([cubes_idxs[cubes_to[i]][0]]).flatten() for idx2 in
                            np.array([cubes_idxs[cubes_to[i]][1]]).flatten()]
                idxs_old = [[idx1, idx2] for idx1 in np.array([cubes_idxs[cubes_from[i]][0]]).flatten() for idx2 in
                            np.array([cubes_idxs[cubes_from[i]][1]]).flatten()]
                for idxNew, idxOld in zip(idxs_new, idxs_old):
                    flat_idx_new: int = int(np.ravel_multi_index((face, idxNew[0], idxNew[1]), colors_new.shape))
                    flat_idx_old: int = int(np.ravel_multi_index((face, idxOld[0], idxOld[1]), colors.shape))
                    rotate_idxs_new[move] = np.concatenate((rotate_idxs_new[move], [flat_idx_new]))
                    rotate_idxs_old[move] = np.concatenate((rotate_idxs_old[move], [flat_idx_old]))

            # Rotate adjacent faces
            face_idxs = adj_idxs[face]
            for i in range(0, len(faces_to)):
                face_to: int = int(faces_to[i])
                face_from: int = int(faces_from[i])
                idxs_new = [[idx1, idx2] for idx1 in np.array([face_idxs[face_to][0]]).flatten() for idx2 in
                            np.array([face_idxs[face_to][1]]).flatten()]
                idxs_old = [[idx1, idx2] for idx1 in np.array([face_idxs[face_from][0]]).flatten() for idx2 in
                            np.array([face_idxs[face_from][1]]).flatten()]
                for idxNew, idxOld in zip(idxs_new, idxs_old):
                    flat_idx_new = int(np.ravel_multi_index((face_to, idxNew[0], idxNew[1]), colors_new.shape))
                    flat_idx_old = int(np.ravel_multi_index((face_from, idxOld[0], idxOld[1]), colors.shape))
                    rotate_idxs_new[move] = np.concatenate((rotate_idxs_new[move], [flat_idx_new]))
                    rotate_idxs_old[move] = np.concatenate((rotate_idxs_old[move], [flat_idx_old]))

        return rotate_idxs_new, rotate_idxs_old

    def __repr__(self) -> str:
        return "Cube3()"



from typing import List, Tuple, Dict, Optional, Any, cast
import numpy as np

from deepxube.base.factory import Parser
from deepxube.base.nnet_input import HasFlatSGAIn, HasFlatSGActsEnumFixedIn, HasTwoDSGActsEnumFixedIn
from deepxube.base.domain import State, Action, Goal, GoalStartRevWalkableActsRev, NextStateNPActsEnumFixed
from deepxube.factories.domain_factory import domain_factory

from numpy.typing import NDArray


class LOState(State):
    __slots__ = ['tiles', 'hash']

    def __init__(self, tiles: NDArray[np.uint8]):
        self.tiles: NDArray[np.uint8] = tiles
        self.hash: Optional[int] = None

    def __hash__(self) -> int:
        if self.hash is None:
            self.hash = hash(self.tiles.tobytes())

        return self.hash

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LOState):
            return np.array_equal(self.tiles, other.tiles)
        return NotImplemented


class LOGoal(Goal):
    def __init__(self, tiles: NDArray[np.uint8]):
        self.tiles: NDArray[np.uint8] = tiles


class LOAction(Action):
    def __init__(self, action: int):
        self.action = action

    def __hash__(self) -> int:
        return self.action

    def __eq__(self, other: object) -> bool:
        if isinstance(other, LOAction):
            return self.action == other.action
        return NotImplemented


@domain_factory.register_class("lightsout")
class LightsOut(NextStateNPActsEnumFixed[LOState, LOAction, LOGoal], GoalStartRevWalkableActsRev[LOState, LOAction, LOGoal],
                HasFlatSGActsEnumFixedIn[LOState, LOAction, LOGoal], HasFlatSGAIn[LOState, LOAction, LOGoal],
                HasTwoDSGActsEnumFixedIn[LOState, LOAction, LOGoal]):
    def __init__(self, dim: int = 7):
        super().__init__()
        self.dim: int = dim
        self.num_tiles: int = self.dim ** 2

        self.move_matrix: NDArray = np.zeros((self.num_tiles, 5), dtype=np.int64)
        for move in range(self.num_tiles):
            x_pos = int(np.floor(move / self.dim))
            y_pos = move % self.dim

            right = move + self.dim if x_pos < (self.dim-1) else move
            left = move - self.dim if x_pos > 0 else move
            up = move + 1 if y_pos < (self.dim - 1) else move
            down = move - 1 if y_pos > 0 else move

            self.move_matrix[move] = [move, right, left, up, down]

        self.actions_fixed: List[LOAction] = [LOAction(x) for x in range(self.num_tiles)]
        self.goal_np: NDArray = np.zeros(self.num_tiles, dtype=np.uint8)

    def is_solved(self, states: List[LOState], goals: List[LOGoal]) -> List[bool]:
        states_np = np.stack([state.tiles for state in states], axis=0)
        goals_np = np.stack([goal.tiles for goal in goals], axis=0)

        return cast(List[bool], np.all(states_np == goals_np, axis=1).tolist())

    def sample_goalstate_goal_pairs(self, num: int) -> Tuple[List[LOState], List[LOGoal]]:
        states_goal: List[LOState] = [LOState(self.goal_np.copy())] * num
        goals: List[LOGoal] = [LOGoal(self.goal_np.copy())] * num

        return states_goal, goals

    def rev_action(self, states: List[LOState], actions: List[LOAction]) -> List[LOAction]:
        return actions

    def get_input_info_flat_sg(self) -> Tuple[List[int], List[int]]:
        return [self.num_tiles], [1]

    def get_input_info_flat_sga(self) -> Tuple[List[int], List[int]]:
        return [self.num_tiles, 1], [1, self.get_num_acts()]

    def to_np_flat_sg(self, states: List[LOState], goals: List[LOGoal]) -> List[NDArray]:
        return [np.stack([x.tiles for x in states], axis=0).astype(np.uint8)]

    def to_np_flat_sga(self, states: List[LOState], goals: List[LOGoal], actions: List[LOAction]) -> List[NDArray]:
        return self.to_np_flat_sg(states, goals) + [np.expand_dims(np.array(self.actions_to_indices(actions)), 1)]

    def get_input_info_2d_sg(self) -> Tuple[List[int], Tuple[int, int], List[int], Optional[int]]:
        return [1], (self.dim, self.dim), [1], 1

    def to_np_2d_sg(self, states: List[LOState], goals: List[LOGoal]) -> List[NDArray]:
        tiles_flat: NDArray[np.uint8] = np.stack([x.tiles for x in states], axis=0).astype(np.uint8)
        return [tiles_flat.reshape((-1, 1, self.dim, self.dim))]

    def actions_to_indices(self, actions: List[LOAction]) -> List[int]:
        return [action_lo.action for action_lo in actions]

    def get_actions_fixed(self) -> List[LOAction]:
        return self.actions_fixed.copy()

    def _states_to_np(self, states: List[LOState]) -> List[NDArray[np.uint8]]:
        return [np.stack([x.tiles for x in states], axis=0)]

    def _np_to_states(self, states_np: List[NDArray]) -> List[LOState]:
        assert len(states_np) == 1
        return [LOState(x) for x in states_np[0]]

    def _next_state_np(self, states_np_l: List[NDArray], actions: List[LOAction]) -> Tuple[List[NDArray], List[float]]:
        assert len(states_np_l) == 1
        tiles_next_np: NDArray = states_np_l[0].copy()

        state_idxs: NDArray = np.arange(0, tiles_next_np.shape[0])
        state_idxs = np.expand_dims(state_idxs, 1)

        actions_np: NDArray = np.array([action.action for action in actions])
        move_matrix = self.move_matrix[actions_np]
        tiles_next_np[state_idxs, move_matrix] = (tiles_next_np[state_idxs, move_matrix] + 1) % 2

        return [tiles_next_np], [1.0] * len(actions)

    def __repr__(self) -> str:
        return f"LightsOut(dim={self.dim})"


@domain_factory.register_parser("lightsout")
class LightsOutParser(Parser):
    def parse(self, args_str: str) -> Dict[str, Any]:
        return {"dim": int(args_str)}

    def help(self) -> str:
        return "An integer for the dimension. E.g. '7'"



from abc import ABC, abstractmethod
from typing import Deque, Tuple, List, Optional, Generic, TypeVar
from collections import deque
from deepxube.base.domain import State, Action, Goal
import numpy as np


ReplayVElem = Tuple[State, Goal, bool]
ReplayQElem = Tuple[State, Goal, bool, Action, float, State]
ReplayPElem = Tuple[State, Goal, Action]

ReplayVRet = Tuple[List[State], List[Goal], List[bool]]
ReplayQRet = Tuple[List[State], List[Goal], List[bool], List[Action], List[float], List[State]]
ReplayPRet = Tuple[List[State], List[Goal], List[Action]]

Elem = TypeVar('Elem')
SampRet = TypeVar('SampRet')


class ReplayBuffer(Generic[Elem, SampRet], ABC):
    def __init__(self, max_size: int):
        self.deque: Deque[Elem] = deque([], max_size)

    def add(self, data: List[Elem]) -> None:
        self.deque.extend(data)

    def sample(self, num: int) -> SampRet:
        assert self.size() > 0, f"Replay buffer size should not be {self.size()}"
        idxs: List[int] = np.random.randint(0, len(self.deque), size=num).tolist()
        elems: List[Elem] = [self.deque[idx] for idx in idxs]
        return self._elems_to_ret(elems)

    def size(self) -> int:
        return len(self.deque)

    def max_size(self) -> int:
        maxlen: Optional[int] = self.deque.maxlen
        assert maxlen is not None

        return maxlen

    @abstractmethod
    def _elems_to_ret(self, elems: List[Elem]) -> SampRet:
        pass


class ReplayBufferV(ReplayBuffer[ReplayVElem, ReplayVRet]):
    def _elems_to_ret(self, elems: List[ReplayVElem]) -> ReplayVRet:
        states: List[State] = [replay_q_elem[0] for replay_q_elem in elems]
        goals: List[Goal] = [replay_q_elem[1] for replay_q_elem in elems]
        is_solved_l: List[bool] = [replay_q_elem[2] for replay_q_elem in elems]

        return states, goals, is_solved_l


class ReplayBufferQ(ReplayBuffer[ReplayQElem, ReplayQRet]):
    def _elems_to_ret(self, elems: List[ReplayQElem]) -> ReplayQRet:
        states: List[State] = [replay_q_elem[0] for replay_q_elem in elems]
        goals: List[Goal] = [replay_q_elem[1] for replay_q_elem in elems]
        is_solved_l: List[bool] = [replay_q_elem[2] for replay_q_elem in elems]
        actions: List[Action] = [replay_q_elem[3] for replay_q_elem in elems]
        tcs: List[float] = [replay_q_elem[4] for replay_q_elem in elems]
        states_next: List[State] = [replay_q_elem[5] for replay_q_elem in elems]

        return states, goals, is_solved_l, actions, tcs, states_next


class ReplayBufferP(ReplayBuffer[ReplayPElem, ReplayPRet]):
    def _elems_to_ret(self, elems: List[ReplayPElem]) -> ReplayPRet:
        states: List[State] = [replay_q_elem[0] for replay_q_elem in elems]
        goals: List[Goal] = [replay_q_elem[1] for replay_q_elem in elems]
        actions: List[Action] = [replay_q_elem[2] for replay_q_elem in elems]

        return states, goals, actions



from abc import ABC, abstractmethod
from typing import List, Tuple, Optional, Set, TypeVar, Generic, Dict, Any
import numpy as np
from clingo.solving import Model as ModelCl

from deepxube.logic.logic_objects import Atom, Model
from deepxube.utils import misc_utils
from deepxube.nnet.nnet_utils import NNetPar, NNetCallable
from deepxube.utils.timing_utils import Times
from matplotlib.figure import Figure
import random
import time
from numpy.typing import NDArray


class State(ABC):
    """ State object

    """
    @abstractmethod
    def __hash__(self) -> int:
        """ For use in CLOSED dictionary for pathfinding
        :return: hash value
        """
        pass

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """ for use in state reidentification during pathfinding

        :param other: other state
        :return: true if they are equal
        """
        pass


class Action(ABC):
    """ Action object

    """

    @abstractmethod
    def __hash__(self) -> int:
        """ For use in backup for Q* search
        :return: hash value
        """
        pass

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """ for use in backup for Q* search

        :param other: other state
        :return: true if they are equal
        """
        pass


class Goal(ABC):
    """ Goal object that represents a set of states considered goal states

    """
    pass


S = TypeVar('S', bound=State)
A = TypeVar('A', bound=Action)
G = TypeVar('G', bound=Goal)


# TODO method for downloading data?
class Domain(ABC, Generic[S, A, G]):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self.nnet_pars: List[Tuple[str, str, NNetPar]] = []

    @abstractmethod
    def sample_problem_instances(self, num_steps_l: List[int], times: Optional[Times] = None) -> Tuple[List[S], List[G]]:
        """ Return start goal pairs with num_steps_l between start and goal

        :param num_steps_l: Number of steps to take between start and goal. Does not have to directly correspond to number of steps and can also be ignored.
        :param times: Times that can be used to profile code
        :return: List of start states and list of goals
        """
        pass

    @abstractmethod
    def sample_state_action(self, states: List[S]) -> List[A]:
        """ Get a random action that is applicable to the current state

        :param states: List of states
        :return: List of random actions applicable to given states
        """
        pass

    @abstractmethod
    def next_state(self, states: List[S], actions: List[A]) -> Tuple[List[S], List[float]]:
        """ Get the next state and transition cost given the current state and action

        :param states: List of states
        :param actions: List of actions to take
        :return: Next states, transition costs
        """
        pass

    @abstractmethod
    def is_solved(self, states: List[S], goals: List[G]) -> List[bool]:
        """ Returns true if the state is a member of the set of goal states represented by the goal

        :param states: List of states
        :param goals: List of goals
        :return: List of booleans where the element at index i corresponds to whether or not the state at index i is a member of the set of goal states
        represented by the goal at index i
        """
        pass

    def sample_next_state(self, states: List[S]) -> Tuple[List[S], List[float]]:
        """ Get random next state and transition cost given the current state

        :param states: List of states
        :return: Next states, transition costs
        """
        actions_rand: List[A] = self.sample_state_action(states)
        return self.next_state(states, actions_rand)

    def random_walk(self, states: List[S], num_steps_l: List[int]) -> Tuple[List[S], List[float]]:
        """ Perform a random walk on the given states for the given number of steps

        :param states: List of states
        :param num_steps_l: number of steps to take for each state
        :return: The resulting state and the path cost for each random walk
        """
        states_walk: List[S] = [state for state in states]
        path_costs: List[float] = [0.0 for _ in states]

        num_steps: NDArray[np.int_] = np.array(num_steps_l)
        num_steps_curr: NDArray[np.int_] = np.zeros(len(states), dtype=int)
        steps_lt: NDArray[np.bool_] = num_steps_curr < num_steps
        while np.any(steps_lt):
            idxs: NDArray[np.int_] = np.where(steps_lt)[0]
            states_to_move = [states_walk[idx] for idx in idxs]

            states_moved, tcs = self.sample_next_state(states_to_move)

            idx: int
            for move_idx, idx in enumerate(idxs):
                states_walk[idx] = states_moved[move_idx]
                path_costs[idx] += tcs[move_idx]

            num_steps_curr[idxs] = num_steps_curr[idxs] + 1

            steps_lt[idxs] = num_steps_curr[idxs] < num_steps[idxs]

        return states_walk, path_costs

    def get_nnet_pars(self) -> List[Tuple[str, str, NNetPar]]:
        return self.nnet_pars

    def set_nnet_fns(self, nnet_fn_dict: Dict[str, NNetCallable]) -> None:
        pass


# Visualization mixins
class StateGoalVizable(Domain[S, A, G]):
    """ Can visualize problem instances

    """
    @abstractmethod
    def visualize_state_goal(self, state: S, goal: G, fig: Figure) -> None:
        pass


class StringToAct(Domain[S, A, G]):
    """ Can get an action from a string. Used when visualizing problem instances.

    """
    @abstractmethod
    def string_to_action(self, act_str: str) -> Optional[A]:
        """
        :param act_str: A string representation of an action
        :return: The action represented by the string, if it is a valid representation, None otherwise
        """
        pass

    @abstractmethod
    def string_to_action_help(self) -> str:
        """
        :return: A description of how actions are represented as strings
        """
        pass


# Action mixins
class ActsFixed(Domain[S, A, G]):
    @abstractmethod
    def sample_action(self, num: int) -> List[A]:
        pass

    def sample_state_action(self, states: List[S]) -> List[A]:
        return self.sample_action(len(states))


class ActsRev(Domain[S, A, G], ABC):
    """ Actions are reversible.

    """
    @abstractmethod
    def rev_action(self, states: List[S], actions: List[A]) -> List[A]:
        """ Get the reverse of the given action

        :param states: List of states
        :param actions: List of actions
        :return: Reverse of given action in the given state
        """
        pass


class ActsEnum(Domain[S, A, G]):
    @abstractmethod
    def get_state_actions(self, states: List[S]) -> List[List[A]]:
        """ Get actions applicable to each states

        :param states: List of states
        :return: Applicable actions
        """
        pass

    def sample_state_action(self, states: List[S]) -> List[A]:
        state_actions_l: List[List[A]] = self.get_state_actions(states)
        return [random.choice(state_actions) for state_actions in state_actions_l]

    def expand(self, states: List[S]) -> Tuple[List[List[S]], List[List[A]], List[List[float]]]:
        """ Generate all children for the state, assumes there is at least one child state
        :param states: List of states
        :return: Children of each state, actions, transition costs for each state
        """
        # TODO further validate
        # initialize
        states_exp_l: List[List[S]] = [[] for _ in range(len(states))]
        actions_exp_l: List[List[A]] = [[] for _ in range(len(states))]
        tcs_l: List[List[float]] = [[] for _ in range(len(states))]
        state_actions: List[List[A]] = self.get_state_actions(states)

        num_actions_tot: NDArray[np.int_] = np.array([len(x) for x in state_actions])
        num_actions_taken: NDArray[np.int_] = np.zeros(len(states), dtype=int)
        actions_lt: NDArray[np.bool_] = num_actions_taken < num_actions_tot

        # for each move, get next states, transition costs, and if solved
        while np.any(actions_lt):
            idxs: NDArray[np.int_] = np.where(actions_lt)[0]
            states_idxs: List[S] = [states[idx] for idx in idxs]
            actions_idxs: List[A] = [state_actions[idx].pop(0) for idx in idxs]

            # next state
            states_next, tcs_move = self.next_state(states_idxs, actions_idxs)

            # append
            idx: int
            for exp_idx, idx in enumerate(idxs):
                states_exp_l[idx].append(states_next[exp_idx])
                actions_exp_l[idx].append(actions_idxs[exp_idx])
                tcs_l[idx].append(tcs_move[exp_idx])

            num_actions_taken[idxs] = num_actions_taken[idxs] + 1
            actions_lt[idxs] = num_actions_taken[idxs] < num_actions_tot[idxs]

        return states_exp_l, actions_exp_l, tcs_l


class ActsEnumFixed(ActsEnum[S, A, G], ActsFixed[S, A, G]):
    def sample_action(self, num: int) -> List[A]:
        actions_fixed: List[A] = self.get_actions_fixed()
        return [random.choice(actions_fixed) for _ in range(num)]

    def get_state_actions(self, states: List[S]) -> List[List[A]]:
        return [self.get_actions_fixed().copy() for _ in range(len(states))]

    @abstractmethod
    def get_actions_fixed(self) -> List[A]:
        pass

    def get_num_acts(self) -> int:
        return len(self.get_actions_fixed())


# Goal mixins
class GoalSampleable(Domain[S, A, G]):
    """ Can sample goals """
    @abstractmethod
    def sample_goals(self, num: int) -> List[G]:
        """ Sample goals
        :return: Goals
        """
        pass


class GoalStateSampleable(Domain[S, A, G]):
    """ Can sample goal states """
    @abstractmethod
    def sample_goal_states(self, num: int) -> List[S]:
        """ Sample goal states
        :return: Goal states
        """
        pass


class GoalSampleableFromState(Domain[S, A, G]):
    """ Can sample goals from states such that the state is a member of the sampled goal """
    @abstractmethod
    def sample_goal_from_state(self, states_start: Optional[List[S]], states_goal: List[S]) -> List[G]:
        """ Given a state, sample a goal that represents a set of goal states of which the given state is a member.

        :param states_start: Optional list of start states. Can be used to sample goals that are difficult to achieve from the given start state.
        :param states_goal: List of states from which goals will be sampled.
        :return: Goals
        """
        pass


class StateSampleableFromGoal(Domain[S, A, G]):
    """ Can sample states from goals """
    @abstractmethod
    def sample_state_from_goal(self, goals: List[G]) -> List[S]:
        """ Given a goal, sample a state that is a member of the set of states represneted by that goal.

        :param goals: List of goals
        :return: List of list of states, where each state is a member of the set of goal states represented by the corresponding goal
        """
        pass


class GoalFixed(GoalSampleable[S, A, G]):
    """ Goal is the same for all problem instances """

    @abstractmethod
    def get_goal(self) -> G:
        """
        :return: Fixed goal
        """
        pass

    def sample_goals(self, num: int) -> List[G]:
        return [self.get_goal()] * num


class GoalStateGoalPairSampleable(Domain[S, A, G]):
    """ Can sample pairs of states and corresponding goals of which the sampled state is a member """
    @abstractmethod
    def sample_goalstate_goal_pairs(self, num: int) -> Tuple[List[S], List[G]]:
        pass


class GoalStateSampGoalSamp(GoalStateGoalPairSampleable[S, A, G], GoalStateSampleable[S, A, G], GoalSampleableFromState[S, A, G], ABC):
    """ Sample goal state and then sample goals from goal states """
    def sample_goalstate_goal_pairs(self, num: int) -> Tuple[List[S], List[G]]:
        states_goal: List[S] = self.sample_goal_states(num)
        goals: List[G] = self.sample_goal_from_state(None, states_goal)
        return states_goal, goals


class GoalSampGoalStateSamp(GoalStateGoalPairSampleable[S, A, G], GoalSampleable[S, A, G], StateSampleableFromGoal[S, A, G], ABC):
    """ Sample goals and then sample goal states from goals """
    def sample_goalstate_goal_pairs(self, num: int) -> Tuple[List[S], List[G]]:
        goals: List[G] = self.sample_goals(num)
        states_goal: List[S] = self.sample_state_from_goal(goals)
        return states_goal, goals


# Problem instance generation mixins
class StartGoalWalkable(GoalSampleableFromState[S, A, G]):
    """ Can sample start states, take actions to obtain another state, and sample a goal from that state"""
    @abstractmethod
    def sample_start_states(self, num_states: int) -> List[S]:
        """ A method for generating start states. Should try to make this generate states that are as diverse as
        possible so that the trained heuristic function generalizes well.

        :param num_states: Number of states to get
        :return: Generated states
        """
        pass

    def sample_problem_instances(self, num_steps_l: List[int], times: Optional[Times] = None) -> Tuple[List[S], List[G]]:
        # Initialize
        if times is None:
            times = Times()

        # Start states
        start_time = time.time()
        states_start: List[S] = self.sample_start_states(len(num_steps_l))
        times.record_time("sample_start_states", time.time() - start_time)

        # random walk
        start_time = time.time()
        states_goal: List[S] = self.random_walk(states_start, num_steps_l)[0]
        times.record_time("random_walk", time.time() - start_time)

        # state to goal
        start_time = time.time()
        goals: List[G] = self.sample_goal_from_state(states_start, states_goal)
        times.record_time("sample_goal", time.time() - start_time)

        return states_start, goals


class GoalStartRevWalkable(GoalStateGoalPairSampleable[S, A, G]):
    def sample_problem_instances(self, num_steps_l: List[int], times: Optional[Times] = None) -> Tuple[List[S], List[G]]:
        # Initialize
        if times is None:
            times = Times()

        # goals
        start_time = time.time()
        states_goal, goals = self.sample_goalstate_goal_pairs(len(num_steps_l))
        times.record_time("sample_goalstate_goal_pairs", time.time() - start_time)

        # random walk to get start states
        start_time = time.time()
        states_start: List[S] = self.random_walk_rev(states_goal, num_steps_l)
        times.record_time("random_walk", time.time() - start_time)

        return states_start, goals

    @abstractmethod
    def random_walk_rev(self, states: List[S], num_steps_l: List[int]) -> List[S]:
        """ Domain need not be reversible as the distance of a path obtained by a number reverse steps can be roughly correlated with the number of steps

        :param states: List of states
        :param num_steps_l: List of integers
        :return: states resulting from reverse random walk
        """
        pass


class GoalStartRevWalkableActsRev(GoalStartRevWalkable[S, A, G], ActsRev[S, A, G], ABC):
    def random_walk_rev(self, states: List[S], num_steps_l: List[int]) -> List[S]:
        return self.random_walk(states, num_steps_l)[0]


# numpy convenience mixins
class NextStateNP(Domain[S, A, G]):
    def next_state(self, states: List[S], actions: List[A]) -> Tuple[List[S], List[float]]:
        states_np: List[NDArray] = self._states_to_np(states)
        states_next_np, tcs = self._next_state_np(states_np, actions)
        states_next: List[S] = self._np_to_states(states_next_np)

        return states_next, tcs

    def random_walk(self, states: List[S], num_steps_l: List[int]) -> Tuple[List[S], List[float]]:
        states_np = self._states_to_np(states)
        path_costs: List[float] = [0.0 for _ in states]

        num_steps: NDArray[np.int_] = np.array(num_steps_l)
        num_steps_curr: NDArray[np.int_] = np.zeros(len(states), dtype=int)
        steps_lt: NDArray[np.bool_] = num_steps_curr < num_steps
        while np.any(steps_lt):
            idxs: NDArray[np.int_] = np.where(steps_lt)[0]
            states_np_tomove: List[NDArray] = [states_np_i[idxs] for states_np_i in states_np]
            actions_rand: List[A] = self._sample_state_np_action(states_np_tomove)

            states_moved, tcs = self._next_state_np(states_np_tomove, actions_rand)

            for l_idx in range(len(states_np)):
                states_np[l_idx][idxs] = states_moved[l_idx]
            idx: int
            for act_idx, idx in enumerate(idxs):
                path_costs[idx] += tcs[act_idx]

            num_steps_curr[idxs] = num_steps_curr[idxs] + 1

            steps_lt[idxs] = num_steps_curr[idxs] < num_steps[idxs]

        return self._np_to_states(states_np), path_costs

    @abstractmethod
    def _states_to_np(self, states: List[S]) -> List[NDArray]:
        pass

    @abstractmethod
    def _np_to_states(self, states_np_l: List[NDArray]) -> List[S]:
        pass

    @abstractmethod
    def _sample_state_np_action(self, states_np: List[NDArray]) -> List[A]:
        pass

    @abstractmethod
    def _next_state_np(self, states_np: List[NDArray], actions: List[A]) -> Tuple[List[NDArray], List[float]]:
        """ Get the next state and transition cost given the current numpy representations of the state and action


        @param states_np: numpy representation of states. Each row in each element of states_np list represents
        information for a different state. There can be one or more multiple elements in the list for each state.
        This object should not be mutated.
        @param actions: actions
        @return: Numpy representation of next states, transition costs
        """
        pass


class NextStateNPActsEnum(NextStateNP[S, A, G], ActsEnum[S, A, G], ABC):
    def expand(self, states: List[S]) -> Tuple[List[List[S]], List[List[A]], List[List[float]]]:
        # initialize
        states_np: List[NDArray] = self._states_to_np(states)
        states_exp_l: List[List[S]] = [[] for _ in range(len(states))]
        actions_exp_l: List[List[A]] = [[] for _ in range(len(states))]
        tcs_l: List[List[float]] = [[] for _ in range(len(states))]
        state_actions: List[List[A]] = self.get_state_actions(states)

        num_actions_tot: NDArray[np.int_] = np.array([len(x) for x in state_actions])
        num_actions_taken: NDArray[np.int_] = np.zeros(len(states), dtype=int)
        actions_lt: NDArray[np.bool_] = num_actions_taken < num_actions_tot

        # for each move, get next states, transition costs, and if solved
        while np.any(actions_lt):
            idxs: NDArray[np.int_] = np.where(actions_lt)[0]
            states_np_idxs: List[NDArray] = [states_np_i[idxs] for states_np_i in states_np]
            actions_idxs: List[A] = [state_actions[idx].pop(0) for idx in idxs]

            # next state
            states_next_np, tcs_move = self._next_state_np(states_np_idxs, actions_idxs)
            states_next: List[S] = self._np_to_states(states_next_np)

            # append
            idx: int
            for exp_idx, idx in enumerate(idxs):
                states_exp_l[idx].append(states_next[exp_idx])
                actions_exp_l[idx].append(actions_idxs[exp_idx])
                tcs_l[idx].append(tcs_move[exp_idx])

            num_actions_taken[idxs] = num_actions_taken[idxs] + 1
            actions_lt[idxs] = num_actions_taken[idxs] < num_actions_tot[idxs]

        return states_exp_l, actions_exp_l, tcs_l


class NextStateNPActsFixed(NextStateNP[S, A, G], ActsFixed[S, A, G], ABC):
    def _sample_state_np_action(self, states_np: List[NDArray]) -> List[A]:
        return self.sample_action(states_np[0].shape[0])


class NextStateNPActsEnumFixed(NextStateNPActsEnum[S, A, G], ActsEnumFixed[S, A, G], ABC):
    def _get_state_np_actions(self, states_np: List[NDArray]) -> List[List[A]]:
        state_actions: List[A] = self.get_actions_fixed()
        return [state_actions.copy() for _ in range(states_np[0].shape[0])]

    def _sample_state_np_action(self, states_np: List[NDArray]) -> List[A]:
        state_actions_l: List[List[A]] = self._get_state_np_actions(states_np)
        return [random.choice(state_actions) for state_actions in state_actions_l]


# PDDL Mixins
class SupportsPDDL(Domain[S, A, G], ABC):
    @abstractmethod
    def get_pddl_domain(self) -> List[str]:
        pass

    @abstractmethod
    def prob_inst_to_pddl_inst(self, state: S, goal: G) -> List[str]:
        pass

    @abstractmethod
    def pddl_action_to_action(self, pddl_action: str) -> A:
        pass


# Logic mixins
class GoalGrndAtoms(GoalSampleableFromState[S, A, G]):
    @abstractmethod
    def state_to_model(self, states: List[S]) -> List[Model]:
        pass

    @abstractmethod
    def model_to_state(self, models: List[Model]) -> List[S]:
        """ Assumes model is a fully specified state

        :param models:
        :return:
        """
        pass

    @abstractmethod
    def goal_to_model(self, goals: List[G]) -> List[Model]:
        pass

    @abstractmethod
    def model_to_goal(self, models: List[Model]) -> List[G]:
        pass

    def is_solved(self, states: List[S], goals: List[G]) -> List[bool]:
        """ Returns whether or not state is solved

        :param states: List of states
        :param goals: List of goals
        :return: Boolean numpy array where the element at index i corresponds to whether or not the
        state at index i is solved
        """
        models_g: List[Model] = self.goal_to_model(goals)
        is_solved_l: List[bool] = []
        models_s: List[Model] = self.state_to_model(states)
        for model_state, model_goal in zip(models_s, models_g):
            is_solved_l.append(model_goal.issubset(model_state))

        return is_solved_l

    def sample_goal_from_state(self, states_start: Optional[List[S]], states_goal: List[S]) -> List[G]:
        models_g: List[Model] = []

        models_s: List[Model] = self.state_to_model(states_goal)
        keep_probs: NDArray[np.float64] = np.random.rand(len(states_goal))
        for model_s, keep_prob in zip(models_s, keep_probs):
            rand_subset: Set[Atom] = misc_utils.random_subset(model_s, keep_prob)
            models_g.append(frozenset(rand_subset))

        return self.model_to_goal(models_g)

    @abstractmethod
    def get_bk(self) -> List[str]:
        """ get background, each element in list is a line

        :return:
        """
        pass

    @abstractmethod
    def get_ground_atoms(self) -> List[Atom]:
        """ Get all possible ground atoms that can be used to make a state

        :return:
        """
        pass

    @abstractmethod
    def on_model(self, m: ModelCl) -> Model:
        """ Process results from clingo

        :param m:
        :return:
        """
        pass

    @abstractmethod
    def start_state_fixed(self, states: List[S]) -> List[Model]:
        """ Given the start state, what must also be true for the goal state (i.e. immovable walls)

        :param states:
        :return:
        """
        pass



from typing import Generic, List, Optional, Any, Tuple, Callable, TypeVar, Dict, Type, Union

from numpy.typing import NDArray

from deepxube.base.domain import Domain, State, Goal, Action, ActsEnum
from deepxube.base.heuristic import HeurFnV, HeurFnQ, PolicyFn
from deepxube.utils import misc_utils
from deepxube.utils.timing_utils import Times
from dataclasses import dataclass

from abc import ABC, abstractmethod
import numpy as np
import time


# pathfinding data structures

class Node:
    __slots__ = ['state', 'goal', 'path_cost', 'heuristic', 'q_values', 'act_probs', 'is_solved', 'parent_action', 'parent_t_cost', 'parent',
                 'edge_dict', 'backup_val']

    def __init__(self, state: State, goal: Goal, path_cost: float, heuristic: float, q_values: Optional[Tuple[List[Action], List[float]]],
                 is_solved: Optional[bool], parent_action: Optional[Action], parent_t_cost: Optional[float], parent: Optional['Node']):
        self.state: State = state
        self.goal: Goal = goal
        self.path_cost: float = path_cost
        self.heuristic: float = heuristic
        self.q_values: Optional[Tuple[List[Action], List[float]]] = q_values
        self.act_probs: Optional[Tuple[List[Action], List[float]]] = None
        self.is_solved: Optional[bool] = is_solved
        self.parent_action: Optional[Action] = parent_action
        self.parent_t_cost: Optional[float] = parent_t_cost
        self.parent: Optional[Node] = parent
        self.edge_dict: Dict[Action, Tuple[float, Node]] = dict()
        self.backup_val: float = np.inf

    def add_edge(self, action: Action, t_cost: float, node_next: "Node") -> None:
        assert action not in self.edge_dict.keys()
        self.edge_dict[action] = (t_cost, node_next)

    def bellman_backup(self) -> float:
        assert self.is_solved is not None

        if self.is_solved:
            self.backup_val = 0.0
        else:
            if len(self.edge_dict) > 0:
                self.backup_val = min(tc + node_c.heuristic for tc, node_c in self.edge_dict.values())
        return self.backup_val

    def upper_bound_parent_path(self, ctg_ub: float) -> None:
        self.backup_val = min(self.backup_val, ctg_ub)
        if self.parent is not None:
            assert self.parent_t_cost is not None
            self.parent.upper_bound_parent_path(ctg_ub + self.parent_t_cost)

    def tree_backup(self) -> float:
        if (self.is_solved is not None) and self.is_solved:
            self.backup_val = 0.0
        else:
            if len(self.edge_dict) == 0:
                self.backup_val = max(self.heuristic, 0.0)
            else:
                self.backup_val = min(tc + node_c.tree_backup() for tc, node_c in self.edge_dict.values())

        return self.backup_val

    def backup_act(self, action: Action) -> float:
        assert self.is_solved is not None
        if self.is_solved:
            return 0.0
        else:
            tc, node_next = self.edge_dict[action]
            # assert node_next.q_values is not None
            if node_next.backup_val < np.inf:
                return tc + node_next.backup_val
            else:
                return tc + node_next.heuristic

    def get_all_descendants(self) -> List['Node']:
        """ Get all descendants of node (excluding self)

        :return: List of nodes that are descendants
        """
        fifo: List[Node] = [x[1] for x in self.edge_dict.values()]
        descendants: List[Node] = []
        while len(fifo) > 0:
            descendant: Node = fifo.pop(0)
            for _, descendant_c in descendant.edge_dict.values():
                fifo.append(descendant_c)
            descendants.append(descendant)
        return descendants


def get_path(node: Node) -> Tuple[List[State], List[Action], float]:
    """ Gets path from the start state to the goal state associated with the input node

    :param node: goal node
    :return: List of states along path, List of actions in path, path cost
    """
    path: List[State] = []
    actions: List[Action] = []

    parent_node: Node = node
    while parent_node.parent is not None:
        path.append(parent_node.state)

        assert parent_node.parent_action is not None, "parent_action should not be None"
        actions.append(parent_node.parent_action)
        parent_node = parent_node.parent

    path.append(parent_node.state)

    path = path[::-1]
    actions = actions[::-1]

    return path, actions, node.path_cost


class EdgeQ:
    __slots__ = ['node', 'action', 'q_val']

    def __init__(self, node: Node, action: Action, q_val: float):
        self.node: Node = node
        self.action: Action = action
        self.q_val: float = q_val


class Instance(ABC):
    def __init__(self, root_node: Node, inst_info: Any):
        self.root_node: Node = root_node
        self.itr: int = 0  # updater with every pathfinding iteration
        self.num_nodes_generated: int = 0
        self.inst_info: Any = inst_info
        self.goal_node: Optional[Node] = None
        self._nodes_curr: List[Node] = [self.root_node]
        self._nodes_popped: List[Node] = []
        self._edges_popped: List[EdgeQ] = []

    @abstractmethod
    def frontier_size(self) -> int:
        pass

    def get_nodes(self) -> List[Node]:
        return self._nodes_curr.copy()

    def set_next_nodes(self, nodes_next: List[Node]) -> None:
        self._nodes_curr = nodes_next.copy()

    @abstractmethod
    def record_goal(self, nodes: List[Node]) -> None:
        pass

    def add_nodes_popped(self, nodes_popped: List[Node]) -> None:
        self._nodes_popped.extend(nodes_popped)

    def get_nodes_popped(self) -> List[Node]:
        return self._nodes_popped.copy()

    def add_edges_popped(self, edges_popped: List[EdgeQ]) -> None:
        self._edges_popped.extend(edges_popped)

    def get_edges_popped(self) -> List[EdgeQ]:
        return self._edges_popped.copy()

    def has_soln(self) -> bool:
        if self.goal_node is None:
            return False
        else:
            return True

    def path_cost(self) -> float:
        if not self.has_soln():
            return np.inf
        else:
            assert self.goal_node is not None
            return self.goal_node.path_cost

    @abstractmethod
    def finished(self) -> bool:
        pass


@dataclass(frozen=True)
class FNsHeurV:
    heur_fn_v: HeurFnV


@dataclass(frozen=True)
class FNsHeurQ:
    heur_fn_q: HeurFnQ


FNsHeur = Union[FNsHeurV, FNsHeurQ]


@dataclass(frozen=True)
class FNsPolicy:
    policy_fn: PolicyFn


@dataclass(frozen=True)
class FNsHeurVPolicy(FNsPolicy, FNsHeurV):
    pass


@dataclass(frozen=True)
class FNsHeurQPolicy(FNsPolicy, FNsHeurQ):
    pass


I = TypeVar('I', bound=Instance)  # noqa: E741
D = TypeVar('D', bound=Domain)
FNs = TypeVar('FNs')


# pathfinding

class PathFind(Generic[D, FNs, I], ABC):
    @staticmethod
    @abstractmethod
    def domain_type() -> Type[D]:
        pass

    @staticmethod
    @abstractmethod
    def functions_type() -> Type[FNs]:
        pass

    def __init__(self, domain: D, functions: FNs):
        assert isinstance(domain, self.domain_type()), f"Domain {domain} must be an instance of {self.domain_type()}."
        if self.functions_type() is not Any:
            assert isinstance(functions, self.functions_type()), f"Functions {functions} must be an instance of {self.functions_type()}."
        self.domain: D = domain
        self.functions: FNs = functions
        self.instances: List[I] = []
        self.times: Times = Times()
        self.itr: int = 0

    @abstractmethod
    def make_instances(self, states: List[State], goals: List[Goal], inst_infos: Optional[List[Any]] = None, compute_root_vals: bool = True) -> List[I]:
        """ Make instances from states and goals

        :param states: List of states
        :param goals: List of goals
        :param inst_infos: Optional list of information to add to an instance
        :param compute_root_vals: If true, compute the values for the root node. Some algorithms may have to ignore this argument and always compute it.
        :return: List of instances
        """
        pass

    def add_instances(self, instances: List[I]) -> None:
        self.instances.extend(instances)

    @abstractmethod
    def expand_states(self, states: List[State], goals: List[Goal]) -> Tuple[List[List[State]], List[List[Action]], List[List[float]]]:
        pass

    @abstractmethod
    def get_state_actions(self, states: List[State], goals: List[Goal]) -> List[List[Action]]:
        pass

    @abstractmethod
    def step(self, verbose: bool = False) -> Tuple[List[Node], List[EdgeQ]]:
        pass

    def remove_finished_instances(self, itr_max: int) -> List[I]:
        def remove_instance_fn(inst_in: I) -> bool:
            if inst_in.finished():
                return True
            if inst_in.itr >= itr_max:
                return True
            return False

        return self.remove_instances(remove_instance_fn)

    def remove_instances(self, test_rem: Callable[[I], bool]) -> List[I]:
        """ Remove instances

        :param test_rem: A Callable that takes an instance as input and returns true if the instance should be removed
        :return: List of removed instances
        """
        instances_remove: List[I] = []
        instances_keep: List[I] = []
        for instance in self.instances:
            if test_rem(instance):
                instances_remove.append(instance)
            else:
                instances_keep.append(instance)

        self.instances = instances_keep

        return instances_remove

    def set_is_solved(self, nodes: List[Node]) -> None:
        start_time = time.time()
        states: List[State] = []
        goals: List[Goal] = []
        for node in nodes:
            states.append(node.state)
            goals.append(node.goal)

        is_solved_l: List[bool] = self.domain.is_solved(states, goals)
        for node, is_solved in zip(nodes, is_solved_l, strict=True):
            node.is_solved = is_solved

        self.times.record_time("is_solved", time.time() - start_time)

    @abstractmethod
    def _set_node_vals(self, nodes: List[Node]) -> None:
        pass

    def _create_root_nodes(self, states: List[State], goals: List[Goal], compute_root_vals: bool) -> List[Node]:
        start_time = time.time()

        root_nodes: List[Node] = []
        for state, goal in zip(states, goals, strict=True):
            root_node: Node = Node(state, goal, 0.0, 0.0, None, None, None, None, None)
            root_nodes.append(root_node)

        self.times.record_time("root", time.time() - start_time)

        if compute_root_vals:
            self._set_node_vals(root_nodes)

        return root_nodes

    def _verbose(self, instances: List[I], nodes_by_inst: List[List[Node]]) -> None:
        nodes_flat: List[Node] = misc_utils.flatten(nodes_by_inst)[0]
        if len(nodes_flat) > 0:
            heuristics: List[float] = [node.heuristic for node in nodes_flat]
            path_costs: List[float] = [node.path_cost for node in nodes_flat]
            frontier_sizes: List[int] = [instance.frontier_size() for instance in instances]
            per_has_soln: float = 100.0 * float(np.mean([inst.has_soln() for inst in instances]))
            per_finished: float = 100.0 * float(np.mean([inst.finished() for inst in instances]))

            print(f"Itr: {self.itr}, Heur(PathCost)(Min/Max): "
                  f"{float(np.min(heuristics)):.2E}({float(path_costs[np.argmin(heuristics)]):.2E})/"
                  f"{float(np.max(heuristics)):.2E}({float(path_costs[np.argmax(heuristics)]):.2E}),"
                  f" Frontier sizes(Min/Max): {min(frontier_sizes)}/{max(frontier_sizes)}, %has_soln: {per_has_soln}, %finished: {per_finished}")
        print(f"Times - {self.times.get_time_str()}\n")


# pathfinding on nodes or edges of graph

class InstanceNode(Instance, ABC):
    @abstractmethod
    def filter_expanded_nodes(self, nodes: List[Node]) -> List[Node]:
        pass

    @abstractmethod
    def push_pop_nodes(self, nodes: List[Node], costs: List[float]) -> List[Node]:
        pass


INode = TypeVar('INode', bound=InstanceNode)


class InstanceEdge(Instance, ABC):
    @abstractmethod
    def filter_popped_nodes(self, nodes: List[Node]) -> List[Node]:
        pass

    @abstractmethod
    def push_pop_edges(self, edges: List[EdgeQ], costs: List[float]) -> List[EdgeQ]:
        pass


IEdge = TypeVar('IEdge', bound=InstanceEdge)


class PathFindNode(PathFind[D, FNs, INode]):
    def step(self, verbose: bool = False) -> Tuple[List[Node], List[EdgeQ]]:
        instances: List[INode] = [instance for instance in self.instances if not instance.finished()]
        if len(instances) == 0:
            self.itr += 1  # TODO make more elegant
            return [], []

        # pop from open
        start_time = time.time()
        nodes_popped_by_inst: List[List[Node]] = [instance.get_nodes() for instance in instances]
        nodes_popped_flat: List[Node] = misc_utils.flatten(nodes_popped_by_inst)[0]
        self.times.record_time("pop", time.time() - start_time)

        # is solved
        self.set_is_solved(nodes_popped_flat)

        # record goal
        start_time = time.time()
        for instance, nodes in zip(instances, nodes_popped_by_inst, strict=True):
            instance.record_goal(nodes)
        self.times.record_time("goal", time.time() - start_time)

        # expand
        nodes_exp_by_inst: List[List[Node]] = self._expand(instances, nodes_popped_by_inst)

        # eval nodes
        self._set_node_vals(misc_utils.flatten(nodes_exp_by_inst)[0])

        # filter expanded nodes
        start_time = time.time()
        for inst_idx, instance in enumerate(instances):
            nodes_exp_by_inst[inst_idx] = instance.filter_expanded_nodes(nodes_exp_by_inst[inst_idx])
        self.times.record_time("filt", time.time() - start_time)

        # get costs
        costs_by_inst: List[List[float]] = self._compute_costs(instances, nodes_exp_by_inst)

        # push
        start_time = time.time()
        nodes_next_by_inst: List[List[Node]] = []
        for instance, nodes_exp, costs in zip(instances, nodes_exp_by_inst, costs_by_inst, strict=True):
            nodes_next_by_inst.append(instance.push_pop_nodes(nodes_exp, costs))
        self.times.record_time("pushpop", time.time() - start_time)

        # get next edges
        start_time = time.time()
        edges_next_flat: List[EdgeQ] = []
        for instance, nodes_next in zip(instances, nodes_next_by_inst):
            edges_popped_inst: List[EdgeQ] = []
            for node in nodes_next:
                assert (node.parent is not None) and (node.parent_action is not None) and (node.parent_t_cost is not None)
                edges_popped_inst.append(EdgeQ(node.parent, node.parent_action, node.parent_t_cost + node.heuristic))
            instance.add_edges_popped(edges_popped_inst)
            edges_next_flat.extend(edges_popped_inst)
        self.times.record_time("edges_next", time.time() - start_time)

        start_time = time.time()
        for instance, nodes_next in zip(instances, nodes_next_by_inst):
            instance.set_next_nodes(nodes_next)
        self.times.record_time("set_next", time.time() - start_time)

        # update iterations
        self.itr += 1
        for instance in instances:
            instance.itr += 1

        # verbose
        if verbose:
            self._verbose(instances, nodes_popped_by_inst)

        return nodes_popped_flat, edges_next_flat

    def _expand(self, instances: List[INode], nodes_by_inst: List[List[Node]]) -> List[List[Node]]:
        start_time = time.time()
        # flatten (for speed)
        nodes: List[Node]
        split_idxs: List[int]
        nodes, split_idxs = misc_utils.flatten(nodes_by_inst)

        if len(nodes) == 0:
            return [[]]

        # Get children of nodes
        states: List[State] = [x.state for x in nodes]
        goals: List[Goal] = [x.goal for x in nodes]

        states_c_l: List[List[State]]
        actions: List[List[Action]]
        tcs: List[List[float]]
        states_c_l, actions, tcs = self.expand_states(states, goals)

        goals_c: List[List[Goal]] = [[node.goal] * len(state_c) for node, state_c in zip(nodes, states_c_l, strict=True)]
        states_c_flat: List[State]
        states_c_flat, split_idxs_c = misc_utils.flatten(states_c_l)
        goals_c_flat, _ = misc_utils.flatten(goals_c)
        self.times.record_time("expand", time.time() - start_time)

        # get children nodes
        start_time = time.time()
        nodes_c: List[Node] = []
        for node_idx, node in enumerate(nodes):
            path_costs_c_i: NDArray = node.path_cost + np.array(tcs[node_idx])
            nodes_c_i: List[Node] = []
            for c_idx in range(len(states_c_l[node_idx])):
                action: Action = actions[node_idx][c_idx]

                edge_dict_val: Optional[Tuple[float, Node]] = node.edge_dict.get(action)
                node_c: Node
                if edge_dict_val is not None:
                    node_c = edge_dict_val[1]
                else:
                    t_cost: float = tcs[node_idx][c_idx]
                    node_c = Node(states_c_l[node_idx][c_idx], goals_c[node_idx][c_idx], float(path_costs_c_i[c_idx]), 0.0, None, None, action, t_cost, node)
                    node.add_edge(action, t_cost, node_c)

                nodes_c_i.append(node_c)
            nodes_c.extend(nodes_c_i)
        self.times.record_time("nodes", time.time() - start_time)

        # get child nodes by instance
        start_time = time.time()
        nodes_c_by_state: List[List[Node]] = misc_utils.unflatten(nodes_c, split_idxs_c)
        nodes_c_by_inst_state: List[List[List[Node]]] = misc_utils.unflatten(nodes_c_by_state, split_idxs)
        nodes_c_by_inst: List[List[Node]] = []
        for nodes_c_by_inst_state_i in nodes_c_by_inst_state:
            nodes_c_by_inst.append(misc_utils.flatten(nodes_c_by_inst_state_i)[0])

        for instance, nodes_by_inst_i, nodes_c_by_inst_i in zip(instances, nodes_by_inst, nodes_c_by_inst, strict=True):
            instance.add_nodes_popped(nodes_by_inst_i)
            instance.num_nodes_generated += len(nodes_c_by_inst_i)

        self.times.record_time("up_inst", time.time() - start_time)

        return nodes_c_by_inst

    @abstractmethod
    def _compute_costs(self, instances: List[INode], nodes_by_inst: List[List[Node]]) -> List[List[float]]:
        pass


class PathFindEdge(PathFind[D, FNs, IEdge]):  # TODO add nodes popped
    def step(self, verbose: bool = False) -> Tuple[List[Node], List[EdgeQ]]:
        instances: List[IEdge] = [instance for instance in self.instances if not instance.finished()]
        if len(instances) == 0:
            self.itr += 1  # TODO make more elegant
            return [], []

        # pop from open
        start_time = time.time()
        nodes_popped_by_inst: List[List[Node]] = [instance.get_nodes() for instance in instances]
        nodes_popped_flat: List[Node] = misc_utils.flatten(nodes_popped_by_inst)[0]
        self.times.record_time("pop", time.time() - start_time)

        # is solved
        self.set_is_solved(nodes_popped_flat)

        # record goal
        start_time = time.time()
        for instance, nodes in zip(instances, nodes_popped_by_inst, strict=True):
            instance.record_goal(nodes)
        self.times.record_time("goal", time.time() - start_time)

        # filter popped nodes
        start_time = time.time()
        for inst_idx, instance in enumerate(instances):
            nodes_popped_by_inst[inst_idx] = instance.filter_popped_nodes(nodes_popped_by_inst[inst_idx])
        self.times.record_time("filt", time.time() - start_time)

        # expand
        edges_exp_by_inst: List[List[EdgeQ]] = self._get_edges(nodes_popped_by_inst)

        # get costs
        costs_by_inst: List[List[float]] = self._compute_costs(instances, edges_exp_by_inst)

        # push
        start_time = time.time()
        edges_next_by_inst: List[List[EdgeQ]] = []
        for instance, edges_exp, costs in zip(instances, edges_exp_by_inst, costs_by_inst, strict=True):
            edges_next_by_inst.append(instance.push_pop_edges(edges_exp, costs))
        self.times.record_time("pushpop", time.time() - start_time)

        # get next nodes
        nodes_next_by_inst: List[List[Node]] = self.get_next_nodes(instances, edges_next_by_inst)

        # eval nodes
        self._set_node_vals(misc_utils.flatten(nodes_next_by_inst)[0])

        start_time = time.time()
        for instance, nodes_next in zip(instances, nodes_next_by_inst):
            instance.set_next_nodes(nodes_next)
        self.times.record_time("set_next", time.time() - start_time)

        # update iterations
        self.itr += 1
        for instance in instances:
            instance.itr += 1

        # verbose
        if verbose:
            self._verbose(instances, nodes_popped_by_inst)

        return nodes_popped_flat, misc_utils.flatten(edges_next_by_inst)[0]

    def get_next_nodes(self, instances: List[IEdge], edges_by_inst: List[List[EdgeQ]]) -> List[List[Node]]:
        if len(instances) == 0:
            return []
        start_time = time.time()
        # flatten
        edges, split_idxs = misc_utils.flatten(edges_by_inst)
        nodes: List[Node] = [edge.node for edge in edges]

        states: List[State] = [node.state for node in nodes]
        goals: List[Goal] = [node.goal for node in nodes]
        path_costs: List[float] = [node.path_cost for node in nodes]
        actions: List[Action] = [edge.action for edge in edges]

        # next states
        states_next, tcs = self.domain.next_state(states, actions)
        path_costs_next: List[float] = (np.array(path_costs) + np.array(tcs)).tolist()
        self.times.record_time("next_state", time.time() - start_time)

        # next nodes
        start_time = time.time()
        nodes_next: List[Node] = []
        for idx in range(len(edges)):
            edge_dict_val: Optional[Tuple[float, Node]] = nodes[idx].edge_dict.get(actions[idx])
            node_next: Node
            if edge_dict_val is not None:
                node_next = edge_dict_val[1]
            else:
                node_next = Node(states_next[idx], goals[idx], path_costs_next[idx], 0.0, None, None, actions[idx], tcs[idx], nodes[idx])
                nodes[idx].add_edge(actions[idx], tcs[idx], node_next)
            nodes_next.append(node_next)
        self.times.record_time("nodes", time.time() - start_time)

        # update instances
        start_time = time.time()
        nodes_next_by_inst: List[List[Node]] = misc_utils.unflatten(nodes_next, split_idxs)
        for instance, edges_by_inst_i, nodes_next_by_inst_i in zip(instances, edges_by_inst, nodes_next_by_inst, strict=True):
            instance.add_edges_popped(edges_by_inst_i)
            instance.num_nodes_generated += len(nodes_next_by_inst_i)
        self.times.record_time("up_inst", time.time() - start_time)

        return nodes_next_by_inst

    def _get_edges(self, nodes_by_inst: List[List[Node]]) -> List[List[EdgeQ]]:
        # make edges
        start_time = time.time()
        edges_by_inst: List[List[EdgeQ]] = []

        for nodes in nodes_by_inst:
            edges: List[EdgeQ] = []
            for node in nodes:
                action_vals: Optional[Tuple[List[Action], List[float]]] = None
                if node.q_values is not None:
                    action_vals = node.q_values
                elif node.act_probs is not None:
                    action_vals = node.act_probs

                assert action_vals is not None
                for action, act_val in zip(action_vals[0], action_vals[1], strict=True):
                    edges.append(EdgeQ(node, action, act_val))
            edges_by_inst.append(edges)
        self.times.record_time("edges", time.time() - start_time)

        return edges_by_inst

    @abstractmethod
    def _compute_costs(self, instances: List[IEdge], edges_by_inst: List[List[EdgeQ]]) -> List[List[float]]:
        pass


# pathfinding with functions

FNsP = TypeVar('FNsP', bound=FNsPolicy)
FNsHV = TypeVar('FNsHV', bound=FNsHeurV)
FNsHQ = TypeVar('FNsHQ', bound=FNsHeurQ)


class PathFindSetPolicy(PathFind[D, FNsP, I], ABC):
    def _set_node_vals(self, nodes: List[Node]) -> None:
        start_time = time.time()
        states: List[State] = [node.state for node in nodes]
        goals: List[Goal] = [node.goal for node in nodes]
        actions_l, probs_l = self.functions.policy_fn(self.domain, states, goals)

        assert len(actions_l) == len(probs_l) == len(states) == len(goals), \
            f"{len(actions_l)}, {len(probs_l)}, {len(states)}, {len(goals)}"

        for node, actions, probs in zip(nodes, actions_l, probs_l, strict=True):
            assert len(actions) == len(probs), f"{len(actions)}, {len(probs)}"
            node.act_probs = (actions, probs)

        self.times.record_time("policy", time.time() - start_time)


class PathFindSetHeurV(PathFind[D, FNsHV, I], ABC):
    def _set_node_vals(self, nodes: List[Node]) -> None:
        start_time = time.time()
        states: List[State] = [node.state for node in nodes]
        goals: List[Goal] = [node.goal for node in nodes]

        heuristics: List[float] = self.functions.heur_fn_v(states, goals)

        assert len(heuristics) == len(states) == len(goals), \
            f"{len(heuristics)}, {len(states)}, {len(goals)}"

        for node, heuristic in zip(nodes, heuristics, strict=True):
            node.heuristic = heuristic

        self.times.record_time("heur", time.time() - start_time)


class PathFindSetHeurQ(PathFind[D, FNsHQ, I], ABC):
    def _set_node_vals(self, nodes: List[Node]) -> None:
        start_time = time.time()
        states: List[State] = [node.state for node in nodes]
        goals: List[Goal] = [node.goal for node in nodes]
        actions_l: List[List[Action]] = self.get_state_actions(states, goals)

        self.times.record_time("actions", time.time() - start_time)

        start_time = time.time()
        qvals_l: List[List[float]] = self.functions.heur_fn_q(states, goals, actions_l)
        heuristics: List[float] = [min(x) for x in qvals_l]

        assert len(heuristics) == len(actions_l) == len(qvals_l) == len(states) == len(goals), \
            f"{len(heuristics)}, {len(actions_l)}, {len(qvals_l)}, {len(states)}, {len(goals)}"

        for node, heuristic, actions, qvals in zip(nodes, heuristics, actions_l, qvals_l, strict=True):
            node.heuristic = heuristic
            node.q_values = (actions, qvals)

        self.times.record_time("heur", time.time() - start_time)


# pathfinding with action spaces

DActsEnum = TypeVar('DActsEnum', bound=ActsEnum)


class PathFindActsEnum(PathFind[DActsEnum, FNs, I], ABC):
    def expand_states(self, states: List[State], goals: List[Goal]) -> Tuple[List[List[State]], List[List[Action]], List[List[float]]]:
        return self.domain.expand(states)

    def get_state_actions(self, states: List[State], goals: List[Goal]) -> List[List[Action]]:
        return self.domain.get_state_actions(states)


class PathFindActsPolicy(PathFind[D, FNsP, I], ABC):
    def expand_states(self, states: List[State], goals: List[Goal]) -> Tuple[List[List[State]], List[List[Action]], List[List[float]]]:
        actions_l: List[List[Action]] = self.functions.policy_fn(self.domain, states, goals)[0]

        # repeat states according to actions
        actions_flat, split_idxs = misc_utils.flatten(actions_l)

        states_flat: List[State] = []
        for state, actions in zip(states, actions_l, strict=True):
            states_flat.extend([state] * len(actions))

        assert len(states_flat) == len(actions_flat), f"{len(states_flat)}, {len(actions_flat)}"

        # get next states
        states_exp_flat, tcs_flat = self.domain.next_state(states_flat, actions_flat)

        # unflatten
        states_exp: List[List[State]] = misc_utils.unflatten(states_exp_flat, split_idxs)
        tcs_l: List[List[float]] = misc_utils.unflatten(tcs_flat, split_idxs)

        return states_exp, actions_l, tcs_l

    def get_state_actions(self, states: List[State], goals: List[Goal]) -> List[List[Action]]:
        return self.functions.policy_fn(self.domain, states, goals)[0]


# pathfinding supervised (for training)

class PathFindSup(PathFind[D, Any, I]):
    """ Use the path cost of a random walk as the learning target.
    See Chervov, Alexander, et al. "A Machine Learning Approach That Beats Large Rubik's Cubes." NeurIPS(2025).

    """
    @staticmethod
    def functions_type() -> Type[Any]:
        return Any

    def make_instances(self, states: List[State], goals: List[Goal], inst_infos: Optional[List[Any]] = None, compute_root_vals: bool = True) -> List[I]:
        raise NotImplementedError

    def expand_states(self, states: List[State], goals: List[Goal]) -> Tuple[List[List[State]], List[List[Action]], List[List[float]]]:
        raise NotImplementedError

    def get_state_actions(self, states: List[State], goals: List[Goal]) -> List[List[Action]]:
        raise NotImplementedError

    @abstractmethod
    def make_instances_rw(self, steps_gen: List[int], inst_infos: Optional[List[Any]]) -> List[I]:
        """ Make instances from a random walk

        """
        pass

    def _set_node_vals(self, nodes: List[Node]) -> None:
        raise NotImplementedError



from typing import Optional, List
import argparse
from argparse import ArgumentParser

from deepxube.factories.updater_factory import get_updater

from deepxube.base.domain import State, Goal
from deepxube.base.heuristic import HeurNNetPar, PolicyNNetPar
from deepxube.base.updater import UpArgs, Update, UpdateHeur, UpdatePolicy
from deepxube.base.trainer import TrainArgs
from deepxube.trainers.utils.train_loop import TestArgs, train
from deepxube.utils.command_line_utils import get_domain_from_arg, get_heur_nnet_par_from_arg, get_policy_nnet_par_from_arg
import numpy as np
import pickle


def parser_train(parser: ArgumentParser) -> None:
    parser.add_argument('--domain', type=str, required=True, help="Domain name and arguments.")

    parser.add_argument('--heur', type=str, default=None, help="Heuristic neural network and arguments.")
    parser.add_argument('--heur_type', type=str, default=None, help="V, QFix, QIn. V maps state/goal tuples to cost-to-go. "
                                                                    "QFix maps state/goal tuples to q_values for a fixed action space. "
                                                                    "QIn maps state/goal/action tuples to q_value (can be used in arbitrary action spaces).")

    parser.add_argument('--policy', type=str, default=None, help="Policy neural network and arguments.")
    parser.add_argument('--policy_samp', type=int, default=10, help="")
    parser.add_argument('--policy_rand', type=int, default=5, help="")
    parser.add_argument('--policy_kl', type=float, default=0.1, help="")

    parser.add_argument('--pathfind', type=str, required=True, help="Pathfinding algorithm and arguments. Batch size of any pathfinding algorithm should be 1 "
                                                                    "since updater assumes 1 instance is generated per iteration.")

    parser.add_argument('--dir', type=str, required=True, help="Directory to save neural networks.")
    parser.add_argument('--skip_heur', action='store_true', default=False, help="Set to skip training of heuristic function.")
    parser.add_argument('--skip_policy', action='store_true', default=False, help="Set to skip training of policy.")

    # train args
    train_group = parser.add_argument_group('train')
    train_group.add_argument('--batch_size', type=int, default=1000, help="Batch size.")
    train_group.add_argument('--lr', type=float, default=0.001, help=" Learning rate.")
    train_group.add_argument('--lr_d', type=float, default=0.9999993, help="Learning rate decay.")
    train_group.add_argument('--max_itrs', type=int, default=100000, help="Maximum training iterations.")
    train_group.add_argument('--display', type=int, default=0, help="Display frequency for nnet training.")
    train_group.add_argument('--bal', action='store_true', default=False, help="Set to balance of number of steps to take to generate problem instances based "
                                                                               "on percentage of states solved.")
    train_group.add_argument('--rb', type=int, default=0, help="Number of updates worth of data to keep in replay buffer. If 0 then no replay buffer is used "
                                                               "and training waits for update to finish to get data and randomly sample from that data. "
                                                               "No replay buffer results in faster updates due to not having to use a separate network to "
                                                               "compute the update, but is more susceptible to instability due to shifts in the distribution "
                                                               "of states seen during search.")
    train_group.add_argument('--up_lt', type=float, default=np.inf, help="Loss must be below this threshold for update.")

    # updater args
    update_group = parser.add_argument_group('update')
    update_group.add_argument('--procs', type=int, default=1, help="Number of processes to generate update data.")
    update_group.add_argument('--step_max', type=int, required=True, help="Maximum number of steps to take when generating problem instnaces.")
    update_group.add_argument('--up_itrs', type=int, default=100, help="Number of iterations to check for update.")
    update_group.add_argument('--up_gen_itrs', type=int, default=100, help="Number of iterations for which to generate training data per update check.")
    update_group.add_argument('--search_itrs', type=int, default=1000, help="Number of search iterations to take when generating data.")
    update_group.add_argument('--up_batch_size', type=int, default=100, help="Maximum number of problem instances to generate at a time. Lower if running out "
                                                                             "of memory.")
    update_group.add_argument('--up_nnet_batch_size', type=int, default=20000, help="Maximum number of inputs to give to any nnet at a time during update. "
                                                                                    "Lower if running out of memory.")
    update_group.add_argument('--her', action='store_true', default=False, help="If problem instance not solved during search, do hindsight experience replay "
                                                                                "(HER) by relabeling deepest node in search tree as a goal state and "
                                                                                "sampling a goal from it.")
    update_group.add_argument('--sync_main', action='store_true', default=False, help="Use main nnet to search during update.")
    update_group.add_argument('--up_v', action='store_true', default=False, help="Verbose update.")

    # update heur args
    update_group.add_argument('--backup', type=int, default=1, help="1 for Bellman backup, -1 for limited horizon bellman lookahead (LHBL)")

    # update policy args
    update_group.add_argument('--policy_rand_p', type=float, default=0.0, help="Probability of sampling random actions for training policy "
                                                                               "(to prevent mode collapse)")

    # test args
    test_group = parser.add_argument_group('test')
    test_group.add_argument('--t_file', type=str, default=None, help="File to use when testing.")
    test_group.add_argument('--t_search_itrs', type=int, default=100, help="Number of search iterations when testing.")
    test_group.add_argument('--t_up_freq', type=int, default=10, help="Test every t_up_freq updates.")
    test_group.add_argument('--t_pathfinds', type=str, default="bwas", help="Comma separated list of pathfinding algorithms to use when testing.")
    test_group.add_argument('--t_init', action='store_true', default=False, help="Set for testing before training begins.")

    # other
    parser.add_argument('--debug', action='store_true', default=False, help="Set for debug mode.")
    parser.set_defaults(func=train_cli)


def train_cli(args: argparse.Namespace) -> None:
    # parse domain and heur_nnet
    domain, domain_name = get_domain_from_arg(args.domain)

    # update args
    up_args: UpArgs = UpArgs(args.procs, args.up_itrs, args.step_max, args.search_itrs, ub_heur_solns=False, backup=args.backup,
                             policy_rand_prob=args.policy_rand_p, up_gen_itrs=args.up_gen_itrs, up_batch_size=args.up_batch_size,
                             nnet_batch_size=args.up_nnet_batch_size, sync_main=args.sync_main, v=args.up_v)

    # parse nnets
    heur_nnet_par: Optional[HeurNNetPar] = None
    update_heur: Optional[UpdateHeur] = None
    policy_nnet_par: Optional[PolicyNNetPar] = None
    update_policy: Optional[UpdatePolicy] = None
    if args.heur is not None:
        heur_nnet_par = get_heur_nnet_par_from_arg(domain, domain_name, args.heur, args.heur_type)[0]
        update_ret: Update = get_updater(domain, args.pathfind, up_args, args.her, "heur")
        assert isinstance(update_ret, UpdateHeur)
        update_heur = update_ret
    if args.policy is not None:
        policy_nnet_par = get_policy_nnet_par_from_arg(domain, domain_name, args.policy, args.policy_samp, args.policy_rand)[0]
        update_ret = get_updater(domain, args.pathfind, up_args, args.her, "policy")
        assert isinstance(update_ret, UpdatePolicy)
        update_policy = update_ret

    # train args
    train_args: TrainArgs = TrainArgs(args.batch_size, args.lr, args.lr_d, args.max_itrs, args.bal, rb=args.rb, loss_thresh=args.up_lt,
                                      policy_kl=args.policy_kl, skip_heur=args.skip_heur, skip_policy=args.skip_policy, display=args.display)

    # test args
    test_args: Optional[TestArgs] = None
    if args.t_file is not None:
        data = pickle.load(open(args.t_file, "rb"))
        states: List[State] = data['states']
        goals: List[Goal] = data['goals']
        test_args = TestArgs(states, goals, args.t_search_itrs, args.t_pathfinds.split(","), args.up_nnet_batch_size, args.t_up_freq, args.t_init)

    train(domain, heur_nnet_par, update_heur, policy_nnet_par, update_policy, args.dir, train_args, test_args=test_args, debug=args.debug)



from abc import ABC
from typing import Dict, Tuple, Type, List, Any, Optional

import numpy as np

from deepxube.base.factory import Factory
from deepxube.base.domain import Domain, State, Action, Goal, ActsEnumFixed
from deepxube.base.nnet_input import NNetInput, StateGoalIn, StateGoalActFixIn, StateGoalActIn, PolicyNNetIn
from deepxube.base.heuristic import HeurNNet, PolicyNNet, HeurNNetPar, PolicyNNetPar, HeurNNetParV, HeurNNetParQIn, HeurNNetParQFixOut

from numpy.typing import NDArray

from deepxube.factories.nnet_input_factory import get_domain_nnet_input_keys, get_nnet_input_t


heuristic_factory: Factory[HeurNNet] = Factory[HeurNNet]("HeurNNet")
policy_factory: Factory[PolicyNNet] = Factory[PolicyNNet]("PolicyNNet")


def build_heur_nnet_par(domain: Domain, domain_name: str, nnet_name: str, nnet_kwargs: Dict[str, Any], heur_type: str) -> HeurNNetPar:
    nnet_input_t: Type[NNetInput] = heuristic_factory.get_type(nnet_name).nnet_input_type()
    nnet_input_domain_keys: List[Tuple[str, str]] = get_domain_nnet_input_keys(domain_name)

    for nnet_input_domain_key in nnet_input_domain_keys:
        nnet_input_cls: Type[NNetInput] = get_nnet_input_t(nnet_input_domain_key)
        if heur_type.upper() == "V":
            if issubclass(nnet_input_cls, StateGoalIn) and issubclass(nnet_input_cls, nnet_input_t):
                return HeurNNetParVConcrete(domain, nnet_input_domain_key, nnet_name, nnet_kwargs)
        elif heur_type.upper() == "QFIX":
            assert isinstance(domain, ActsEnumFixed)
            if issubclass(nnet_input_cls, StateGoalActFixIn) and issubclass(nnet_input_cls, nnet_input_t):
                return HeurNNetParQFixOutConcrete(domain, nnet_input_domain_key, nnet_name, nnet_kwargs, domain.get_num_acts())
        elif heur_type.upper() == "QIN":
            if issubclass(nnet_input_cls, StateGoalActIn) and issubclass(nnet_input_cls, nnet_input_t):
                return HeurNNetParQActInConcrete(domain, nnet_input_domain_key, nnet_name, nnet_kwargs)
        else:
            raise ValueError(f"Unknown heur type {heur_type}")
    raise ValueError(f"Cannot build heur nnet for domain: {domain_name}, heur type {heur_type}, and "
                     f"nnet_input type {nnet_input_t}.\nNNet inputs checked: {nnet_input_domain_keys}")


def build_policy_nnet_par(domain: Domain, domain_name: str, nnet_name: str, nnet_kwargs: Dict[str, Any], num_samp: int, num_rand: int) -> PolicyNNetPar:
    nnet_input_t: Type[NNetInput] = policy_factory.get_type(nnet_name).nnet_input_type()
    nnet_input_domain_keys: List[Tuple[str, str]] = get_domain_nnet_input_keys(domain_name)
    for nnet_input_domain_key in nnet_input_domain_keys:
        nnet_input_cls: Type[NNetInput] = get_nnet_input_t(nnet_input_domain_key)
        if issubclass(nnet_input_cls, PolicyNNetIn) and issubclass(nnet_input_cls, nnet_input_t):
            return PolicyNNetParConcrete(domain, nnet_input_domain_key, nnet_name, nnet_kwargs, num_samp, num_rand)

    raise ValueError(f"Cannot build policy nnet for domain: {domain_name}, and "
                     f"nnet_input type {nnet_input_t}.\nNNet inputs checked: {nnet_input_domain_keys}")


class HeurNNetParFacClass(HeurNNetPar, ABC):
    def __init__(self, domain: Domain, nnet_input_name: Tuple[str, str], nnet_name: str, nnet_kwargs: Dict[str, Any], q_fix: bool, out_dim: int):
        self.domain: Domain = domain
        self.nnet_input_name: Tuple[str, str] = nnet_input_name
        self.nnet_input: Optional[NNetInput] = None
        self.nnet_name: str = nnet_name
        self.nnet_kwargs: Dict[str, Any] = nnet_kwargs
        self.q_fix: bool = q_fix
        self.out_dim: int = out_dim

    def get_nnet(self) -> HeurNNet:
        nnet_params: Dict = self.nnet_kwargs.copy()
        nnet_params['nnet_input'] = self._get_nnet_input()
        nnet_params['q_fix'] = self.q_fix
        nnet_params['out_dim'] = self.out_dim
        return heuristic_factory.build_class(self.nnet_name, nnet_params)

    def _get_nnet_input(self) -> NNetInput:
        if self.nnet_input is None:
            self.nnet_input = get_nnet_input_t(self.nnet_input_name)(domain=self.domain)
        return self.nnet_input

    def __getstate__(self) -> Dict:
        self.nnet_input = None
        return self.__dict__


class PolicyNNetParFacClass(PolicyNNetPar, ABC):
    def __init__(self, domain: Domain, nnet_input_name: Tuple[str, str], nnet_name: str, nnet_kwargs: Dict[str, Any], num_samp: int, num_rand: int):
        super().__init__(num_samp, num_rand)
        self.domain: Domain = domain
        self.nnet_input_name: Tuple[str, str] = nnet_input_name
        self.nnet_input: Optional[PolicyNNetIn] = None
        self.nnet_name: str = nnet_name
        self.nnet_kwargs: Dict[str, Any] = nnet_kwargs

    def get_nnet(self) -> PolicyNNet:
        nnet_params: Dict = self.nnet_kwargs.copy()
        nnet_params['nnet_input'] = self._get_nnet_input()
        return policy_factory.build_class(self.nnet_name, nnet_params)

    def _get_nnet_input(self) -> PolicyNNetIn:
        if self.nnet_input is None:
            nnet_input: NNetInput = get_nnet_input_t(self.nnet_input_name)(domain=self.domain)
            assert isinstance(nnet_input, PolicyNNetIn)
            self.nnet_input = nnet_input
        return self.nnet_input

    def __getstate__(self) -> Dict:
        self.nnet_input = None
        return self.__dict__


class HeurNNetParVConcrete(HeurNNetParV, HeurNNetParFacClass):
    def __init__(self, domain: Domain, nnet_input_name: Tuple[str, str], nnet_name: str,
                 nnet_kwargs: Dict[str, Any]):
        HeurNNetParFacClass.__init__(self, domain, nnet_input_name, nnet_name, nnet_kwargs, False, 1)

    def to_np(self, states: List[State], goals: List[Goal]) -> List[NDArray]:
        return self._get_nnet_input().to_np(states, goals)

    def _get_nnet_input(self) -> StateGoalIn:
        nnet_input: NNetInput = super()._get_nnet_input()
        assert isinstance(nnet_input, StateGoalIn)
        return nnet_input


class HeurNNetParQFixOutConcrete(HeurNNetParQFixOut, HeurNNetParFacClass):
    def __init__(self, domain: Domain, nnet_input_name: Tuple[str, str], nnet_name: str, nnet_kwargs: Dict[str, Any], out_dim: int):
        HeurNNetParFacClass.__init__(self, domain, nnet_input_name, nnet_name, nnet_kwargs, True, out_dim)

    def _to_np_fixed_acts(self, states: List[State], goals: List[Goal], actions_l: List[List[Action]]) -> List[NDArray]:
        return self._get_nnet_input().to_np(states, goals, actions_l)

    def _get_nnet_input(self) -> StateGoalActFixIn:
        nnet_input: NNetInput = super()._get_nnet_input()
        assert isinstance(nnet_input, StateGoalActFixIn)
        return nnet_input


class HeurNNetParQActInConcrete(HeurNNetParQIn, HeurNNetParFacClass):
    def __init__(self, domain: Domain, nnet_input_name: Tuple[str, str],
                 nnet_name: str, nnet_kwargs: Dict[str, Any]):
        HeurNNetParFacClass.__init__(self, domain, nnet_input_name, nnet_name, nnet_kwargs, False, 1)

    def _to_np_one_act(self, states: List[State], goals: List[Goal], actions: List[Action]) -> List[NDArray]:
        return self._get_nnet_input().to_np(states, goals, actions)

    def _get_nnet_input(self) -> StateGoalActIn:
        nnet_input: NNetInput = super()._get_nnet_input()
        assert isinstance(nnet_input, StateGoalActIn)
        return nnet_input


class PolicyNNetParConcrete(PolicyNNetParFacClass):
    def to_np_fn(self, states: List[State], goals: List[Goal]) -> List[NDArray[Any]]:
        return self._get_nnet_input().to_np_fn(states, goals)

    def to_np_train(self, states: List[State], goals: List[Goal], actions: List[Action]) -> List[NDArray[Any]]:
        return self._get_nnet_input().to_np(states, goals, actions)

    def _nnet_out_to_actions(self, nnet_out: List[NDArray[np.float64]]) -> List[Action]:
        return self._get_nnet_input().nnet_out_to_actions(nnet_out)



from typing import List, Tuple

from deepxube.base.heuristic import PolicyNNet
from deepxube.base.updater import UpdatePolicy
from deepxube.base.trainer import Train
from deepxube.trainers.utils.train_utils import train_policy_nnet_step
from deepxube.utils.timing_utils import Times

import numpy as np
from numpy.typing import NDArray
import time


class TrainPolicy(Train[PolicyNNet, UpdatePolicy]):
    @staticmethod
    def data_parallel() -> bool:
        return False

    def _train_itr(self, batch: List[NDArray], first_itr_in_update: bool, times: Times) -> float:
        start_time = time.time()

        self.nnet.train()
        loss = train_policy_nnet_step(self.nnet, batch, self.optimizer, self.device, self.status.itr, self.train_args, self.train_start_time)
        self.writer.add_scalar("train/loss", loss, self.status.itr)

        times.record_time("train", time.time() - start_time)
        return loss

    def _add_post_up_info(self) -> List[str]:
        return []

    def _get_shapes_dtypes(self) -> List[Tuple[Tuple[int, ...], np.dtype]]:
        return self.updater.get_policy_train_shapes_dtypes()



from typing import List, Tuple

from deepxube.base.heuristic import HeurNNet
from deepxube.base.updater import UpdateHeur
from deepxube.base.trainer import Train
from deepxube.utils.timing_utils import Times
from deepxube.trainers.utils.train_utils import train_heur_nnet_step, ctgs_summary

import torch.nn as nn

import numpy as np
from numpy.typing import NDArray
import time


class TrainHeur(Train[HeurNNet, UpdateHeur]):
    @staticmethod
    def data_parallel() -> bool:
        return True

    def _train_itr(self, batch: List[NDArray], first_itr_in_update: bool, times: Times) -> float:
        start_time = time.time()
        inputs_batch_np: List[NDArray] = batch[:-1]
        ctgs_batch_np: NDArray = batch[-1]
        ctgs_batch_np = np.expand_dims(ctgs_batch_np.astype(np.float32), 1)

        self.nnet.train()
        ctgs_batch_nnet, loss = train_heur_nnet_step(self.nnet, inputs_batch_np, ctgs_batch_np, self.optimizer, nn.MSELoss(), self.device, self.status.itr,
                                                     self.train_args, self.train_start_time)
        self.writer.add_scalar("train/loss", loss, self.status.itr)

        if first_itr_in_update:
            self.train_summary.itr_to_in_out[self.status.itr] = (ctgs_batch_np, ctgs_batch_nnet)
        times.record_time("train", time.time() - start_time)
        return loss

    def _add_post_up_info(self) -> List[str]:
        ctgs_mean, ctgs_min, ctgs_max = ctgs_summary([self.db.arrays[-1]])
        self.writer.add_scalar("train/ctgs/mean", ctgs_mean, self.status.itr)
        self.writer.add_scalar("train/ctgs/min", ctgs_min, self.status.itr)
        self.writer.add_scalar("train/ctgs/max", ctgs_max, self.status.itr)
        return [f"cost-to-go (mean/min/max): {ctgs_mean:.2f}/{ctgs_min:.2f}/{ctgs_max:.2f}"]

    def _get_shapes_dtypes(self) -> List[Tuple[Tuple[int, ...], np.dtype]]:
        return self.updater.get_heur_train_shapes_dtypes()



from abc import ABC
from typing import List, Any, Optional, Type, TypeVar, Tuple
from deepxube.base.domain import Domain, State, Goal, StartGoalWalkable, GoalStartRevWalkable
from deepxube.base.pathfinding import InstanceNode, Node, EdgeQ, PathFindNode, PathFindSup
from deepxube.factories.pathfinding_factory import pathfinding_factory
import time


class InstanceNodeSup(InstanceNode):
    def filter_expanded_nodes(self, nodes: List[Node]) -> List[Node]:
        raise NotImplementedError

    def frontier_size(self) -> int:
        raise NotImplementedError

    def record_goal(self, nodes: List[Node]) -> None:
        raise NotImplementedError

    def push_pop_nodes(self, nodes: List[Node], costs: List[float]) -> List[Node]:
        raise NotImplementedError

    def __init__(self, root_node: Node, path_cost_sup: float, inst_info: Any):
        super().__init__(root_node, inst_info)
        self.path_cost_sup: float = path_cost_sup

    def finished(self) -> bool:
        return self.itr > 0


D = TypeVar('D', bound=Domain)


class PathFindNodeSup(PathFindNode[D, Any, InstanceNodeSup], PathFindSup[D, InstanceNodeSup], ABC):
    def step(self, verbose: bool = False) -> Tuple[List[Node], List[EdgeQ]]:
        nodes: List[Node] = []
        for instance in self.instances:
            node_root: Node = instance.root_node
            node_root.heuristic = instance.path_cost_sup
            node_root.backup_val = instance.path_cost_sup
            nodes.append(node_root)
            instance.add_nodes_popped([node_root])
            instance.itr += 1
        start_time = time.time()
        self.set_is_solved(nodes)
        self.times.record_time("is_solved", time.time() - start_time)

        return nodes, []

    def _compute_costs(self, instances: List[InstanceNodeSup], nodes_by_inst: List[List[Node]]) -> List[List[float]]:
        raise NotImplementedError

    def _make_instances(self, states_start: List[State], goals: List[Goal], path_costs: List[float], inst_infos: Optional[List[Any]]) -> List[InstanceNodeSup]:
        nodes_root: List[Node] = self._create_root_nodes(states_start, goals, False)

        start_time = time.time()
        if inst_infos is None:
            inst_infos = [None for _ in states_start]

        instances: List[InstanceNodeSup] = []
        for node_root, path_cost, inst_info in zip(nodes_root, path_costs, inst_infos):
            instances.append(InstanceNodeSup(node_root, path_cost, inst_info))
        self.times.record_time("instances", time.time() - start_time)

        return instances


@pathfinding_factory.register_class("sup_v_rw")
class PathFindNodeSupRW(PathFindNodeSup[StartGoalWalkable]):
    @staticmethod
    def domain_type() -> Type[StartGoalWalkable]:
        return StartGoalWalkable

    def make_instances_rw(self, steps_gen: List[int], inst_infos: Optional[List[Any]]) -> List[InstanceNodeSup]:
        start_time = time.time()
        states_start: List[State] = self.domain.sample_start_states(len(steps_gen))
        self.times.record_time("get_start_states", time.time() - start_time)

        start_time = time.time()
        states_goal, path_costs = self.domain.random_walk(states_start, steps_gen)
        self.times.record_time("random_walk", time.time() - start_time)

        # state to goal
        start_time = time.time()
        goals: List[Goal] = self.domain.sample_goal_from_state(states_start, states_goal)
        self.times.record_time("sample_goal", time.time() - start_time)

        return self._make_instances(states_start, goals, path_costs, inst_infos)


@pathfinding_factory.register_class("sup_v_rw_rev")
class PathFindNodeSupRWRev(PathFindNodeSup[GoalStartRevWalkable]):
    @staticmethod
    def domain_type() -> Type[GoalStartRevWalkable]:
        return GoalStartRevWalkable

    def make_instances_rw(self, steps_gen: List[int], inst_infos: Optional[List[Any]]) -> List[InstanceNodeSup]:
        start_time = time.time()
        states_goal, goals = self.domain.sample_goalstate_goal_pairs(len(steps_gen))
        self.times.record_time("samp_goal_state_goal", time.time() - start_time)

        start_time = time.time()
        states_start, path_costs = self.domain.random_walk(states_goal, steps_gen)
        self.times.record_time("random_walk", time.time() - start_time)

        return self._make_instances(states_start, goals, path_costs, inst_infos)



from typing import List, Dict, Any, Type
import torch
from torch import nn, Tensor
import re

from deepxube.base.factory import Parser
from deepxube.base.nnet_input import TwoDIn
from deepxube.base.heuristic import HeurNNet
from deepxube.nnet.pytorch_models import Conv2dModel, ResnetModel, OneHot

from deepxube.factories.heuristic_factory import heuristic_factory


@heuristic_factory.register_class("resnet_2d")
class Resnet2D(HeurNNet[TwoDIn]):
    @staticmethod
    def nnet_input_type() -> Type[TwoDIn]:
        return TwoDIn

    def __init__(self, nnet_input: TwoDIn, out_dim: int, q_fix: bool, num_chan: int = 64, num_blocks: int = 4,
                 batch_norm: bool = False, weight_norm: bool = False, act_fn: str = "RELU"):
        super().__init__(nnet_input, out_dim, q_fix)

        chan_dims, (height, width), one_hot_depths, q_fix_1x1 = self.nnet_input.get_input_info()

        # one hots
        self.one_hots: nn.ModuleList = nn.ModuleList()
        chan_in_tot: int = 0
        for chan_dim, one_hot_depth in zip(chan_dims, one_hot_depths, strict=True):
            assert one_hot_depth >= 1
            self.one_hots.append(OneHot(one_hot_depth, False))
            chan_in_tot += chan_dim * one_hot_depth

        # res net
        def res_block_init() -> nn.Module:
            return Conv2dModel(num_chan, [num_chan] * 2, [3] * 2, [1] * 2, [act_fn, "LINEAR"],
                               batch_norms=[batch_norm] * 2, weight_norms=[weight_norm] * 2)

        self.heur = nn.Sequential(
            Conv2dModel(chan_in_tot, [num_chan], [1], [0], ["LINEAR"]),
            ResnetModel(res_block_init, num_blocks, act_fn),
        )

        if self.q_fix and (q_fix_1x1 is not None):
            assert (height * width * q_fix_1x1) == out_dim
            self.out = nn.Sequential(
                Conv2dModel(num_chan, [q_fix_1x1], [1], [0], ["LINEAR"]),
                nn.Flatten(),
            )
        else:
            self.out = nn.Sequential(
                Conv2dModel(num_chan, [1], [1], [0], ["LINEAR"]),
                nn.Flatten(),
                nn.Linear(height * width, out_dim)
            )

    def _forward(self, inputs: List[Tensor]) -> Tensor:
        inputs_oh: List[Tensor] = []
        for input_i, one_hot in zip(inputs, self.one_hots):
            input_i_oh: Tensor = one_hot(input_i)
            if len(input_i_oh.shape) == 5:
                input_i_oh = input_i_oh.permute((0, 1, 4, 2, 3)).flatten(1, 2)
            inputs_oh.append(input_i_oh)

        # inputs_oh: List[Tensor] = [one_hot(input_i).permute((0, 1, 4, 2, 3)).flatten(1, 2) for input_i, one_hot in zip(inputs, self.one_hots)]
        x: Tensor = self.heur(torch.cat(inputs_oh, dim=1))
        x = self.out(x)
        return x


@heuristic_factory.register_parser("resnet_2d")
class ResnetFCParser(Parser):
    def parse(self, args_str: str) -> Dict[str, Any]:
        args_str_l: List[str] = args_str.split("_")
        kwargs: Dict[str, Any] = dict()
        for args_str_i in args_str_l:
            chan_re = re.search(r"^(\S+)C$", args_str_i)
            blocks_re = re.search(r"^(\S+)B$", args_str_i)
            bn_re = re.search(r"^bn$", args_str_i)
            wn_re = re.search(r"^wn$", args_str_i)
            if chan_re is not None:
                kwargs["num_chan"] = int(chan_re.group(1))
            elif blocks_re is not None:
                kwargs["num_blocks"] = int(blocks_re.group(1))
            elif bn_re is not None:
                kwargs["batch_norm"] = True
            elif wn_re is not None:
                kwargs["weight_norm"] = True
            else:
                raise ValueError(f"Unexpected argument {args_str_i!r}")
        return kwargs

    def help(self) -> str:
        return ("Arguments are delimited by '_' and can be in any order.\n<num>C (number of channels), "
                "<num>B (number of blocks), bn (batch_norm), wn (weight_norm).\n"
                "E.g. resnet_2d.64C_4B_bn")
